"""
Run all examples in `radis/examples <https://github.com/radis/radis/tree/develop/examples>`__
"""

import os
import runpy
import shutil
from os.path import abspath, dirname, exists, join
from pathlib import Path

import pytest

EXAMPLE_FOLDER = abspath(join(dirname(__file__), "..", "..", "examples"))

scripts = [
    join(EXAMPLE_FOLDER, k) for k in os.listdir(EXAMPLE_FOLDER) if k.endswith(".py")
]


@pytest.mark.parametrize("script", scripts)
def test_script_execution(script):
    """Run all examples in `radis/examples <https://github.com/radis/radis/tree/develop/examples>`__"""
    # ensure Matplotlib is interactive (figures won't block the rest of the execution)
    import matplotlib.pyplot as plt

    plt.ion()
    # run:
    runpy.run_path(script, init_globals=locals())
    plt.close("all")

    # Clean
    if exists(Path.cwd() / "SpecDatabase_Test"):
        shutil.rmtree(Path.cwd() / "SpecDatabase_Test")


if __name__ == "__main__":
    print(scripts)
    pytest.main(["test_examples.py", "--pdb"])



# -*- coding: utf-8 -*-
"""Contains the :py:class:`~radis.lbl.factory.SpectrumFactory` class, which is
the core of the RADIS Line-by-Line module.

Examples
--------

Calculate a CO Spectrum, fetching the lines from HITRAN ::

    # This is how you get a spectrum (see spectrum.py for front-end functions
    # that do just that)
    sf = SpectrumFactory(2125, 2249.9,
                         molecule='CO',
                         isotope=1,
                         cutoff=1e-30,   # for faster calculations. See
                                         # `plot_linestrength_hist` for more details
                         **kwargs)
    sf.fetch_databank()        # autodownload from HITRAN
    s = sf.eq_spectrum(Tgas=300)
    s.plot('abscoeff')          # opacity

    # Here we get some extra informations:
    s.plot('radiance', wunit='nm',
                       Iunit='µW/cm2/sr/nm',  # Iunit is arbitrary. Use whatever makes sense
                       show_points=True)  # show_points to have an idea of the resolution

See :py:mod:`radis.test.lbl.test_factory` for more examples


Routine Listing
---------------
*(the following sections are marked on the slider bar in the Spyder IDE with
the # XXX symbol)*

PUBLIC METHODS

- :meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum`             >>> calc equilibrium spectrum
- :meth:`~radis.lbl.factory.SpectrumFactory.non_eq_spectrum`         >>> calc non equilibrium spectrum
- :meth:`~radis.lbl.factory.SpectrumFactory.optically_thin_power`    >>> get total power (equilibrium or non eq)

Most methods are written in inherited class with the following inheritance scheme:

:py:class:`~radis.lbl.loader.DatabankLoader` > :py:class:`~radis.lbl.base.BaseFactory` >
:py:class:`~radis.lbl.broadening.BroadenFactory` > :py:class:`~radis.lbl.bands.BandFactory` >
:py:class:`~radis.lbl.factory.SpectrumFactory`


.. inheritance-diagram:: radis.lbl.factory.SpectrumFactory
   :parts: 1


Notes
-----

Debugging:

Remember one can access all locals variable in a function either by running
%debug after a crash in IPython. Or by adding the line `globals().update(locals())`
at any point


Performance:

Fast version: iterate over chunks of dataframe
Note that we can't use the full dataframe because it's too big and takes too much memory
See :ref:`Performance <label_lbl_performance>` for more details.


for Developers:

- To implement new database formats, see the databases parsers in cdsd.py / hitran.py,
  and the partition function interpolators / calculators methods of SpectrumFactory:
  :py:meth:`~radis.lbl.loader.DatabankLoader._build_partition_function_calculator` and
  :py:meth:`~radis.lbl.loader.DatabankLoader._build_partition_function_interpolator`


----------
"""
from typing import Union
from warnings import warn

import astropy.units as u
import numpy as np
from numpy import arange, exp
from scipy.constants import N_A, c, k, pi
from scipy.optimize import OptimizeResult

from radis import version
from radis.db import MOLECULES_LIST_EQUILIBRIUM, MOLECULES_LIST_NONEQUILIBRIUM
from radis.db.classes import get_molecule, get_molecule_identifier

try:  # Proper import
    from .bands import BandFactory
    from .base import get_waverange
except ImportError:  # if ran from here
    from radis.lbl.bands import BandFactory
    from radis.lbl.base import get_waverange

from radis.misc.basics import flatten, is_float, list_if_float, round_off
from radis.misc.printer import printg
from radis.misc.utils import Default
from radis.phys.constants import k_b
from radis.phys.convert import conv2
from radis.phys.units import convert_emi2nm, convert_rad2nm
from radis.phys.units_astropy import convert_and_strip_units
from radis.spectrum.equations import calc_radiance
from radis.spectrum.spectrum import Spectrum

c_cm = c * 100

# %% Main functions


class SpectrumFactory(BandFactory):
    """A class to put together all functions related to loading CDSD / HITRAN
    databases, calculating the broadenings, and summing over all the lines.

    Parameters
    ----------
    wmin, wmax : ``float`` or `~astropy.units.quantity.Quantity`
        a hybrid parameter which can stand for minimum (maximum) wavenumber or minimum
        (maximum) wavelength depending upon the unit accompanying it. If dimensionless,
        ``wunit`` is considered as the accompanying unit.
    wunit: ``'nm'``, ``'cm-1'``
        the unit accompanying wmin and wmax. Can only be passed with wmin
        and wmax. Default is ``"cm-1"``.
    wavenum_min, wavenum_max: ``float(cm^-1)`` or `~astropy.units.quantity.Quantity`
        minimum (maximum) wavenumber to be processed in :math:`cm^{-1}`.
        use astropy.units to specify arbitrary inverse-length units.
    wavelength_min, wavelength_max : ``float(nm)`` or `~astropy.units.quantity.Quantity`
        minimum (maximum) wavelength to be processed in :math:`nm`. This wavelength
        can be in ``'air'`` or ``'vacuum'`` depending on the value of the parameter
        ``medium=``.
        use astropy.units to specify arbitrary length units.
    pressure: ``float(bar)`` or `~astropy.units.quantity.Quantity`
        partial pressure of gas in bar. Default ``1.01325`` (1 atm).
        use astropy.units to specify arbitrary pressure units.
        For example, ``1013.25 * u.mbar``.
    mole_fraction: ``float``  [ 0 - 1]
        species mole fraction. Default ``1``. Note that the rest of the gas
        is considered to be air for collisional broadening.
    path_length: ``float(cm)`` or `~astropy.units.quantity.Quantity`
        path length in cm. Default ``1``.
        use astropy.units to specify arbitrary length units.
    molecule: ``int``, ``str``, or ``None``
        molecule id (HITRAN format) or name. If ``None``, the molecule can be infered
        from the database files being loaded. See the list of supported molecules
        in :py:data:`~radis.db.MOLECULES_LIST_EQUILIBRIUM`
        and :py:data:`~radis.db.MOLECULES_LIST_NONEQUILIBRIUM`.
        Default ``None``.
    isotope: ``int``, ``list``, ``str`` of the form ``'1,2'``, or ``'all'``
        isotope id (sorted by relative density: (eg: 1: CO2-626, 2: CO2-636 for CO2).
        See HITRAN documentation for isotope list for all species. If 'all',
        all isotopes in database are used (this may result in larger computation
        times!). Default ``'all'``
    medium: ``'air'``, ``'vacuum'``
        propagating medium when giving inputs with ``'wavenum_min'``, ``'wavenum_max'``.
        Does not change anything when giving inputs in wavenumber. Default ``'air'``

    Other Parameters
    ----------------
    Tref: K
        Reference temperature for calculations (linestrength temperature
        correction). HITRAN database uses 296 Kelvin. Default 296 K
    self_absorption: boolean
        Compute self absorption. If ``False``, spectra are optically thin. Default ``True``.
    truncation: float (:math:`cm^{-1}`)
        Half-width over which to compute the lineshape, i.e. lines are truncated
        on each side after ``truncation`` (:math:`cm^{-1}`) from the line center.
        If ``None``, use no truncation (lineshapes spread on the full spectral range).
        Default is ``300`` :math:`cm^{-1}`

        .. note::
         Large values (> ``50``) can induce a performance drop (computation of lineshape
         typically scale as :math:`~truncation ^2` ). The default ``300`` was
         chosen to maintain a good accuracy, and still exhibit the sub-Lorentzian
         behavior of most lines far (few hundreds :math:`cm^{-1}`) from the line center.
    neighbour_lines: float (:math:`cm^{-1}`)
        The calculated spectral range is increased (by ``neighbour_lines`` cm-1
        on each side) to take into account overlaps from out-of-range lines.
        Default is ``0`` :math:`cm^{-1}`.​
    wstep: float (cm-1) or `'auto'`
        Resolution of wavenumber grid. Default ``0.01`` cm-1.
        If `'auto'`, it is ensured that there
        are slightly more points for each linewidth than the value of ``"GRIDPOINTS_PER_LINEWIDTH_WARN_THRESHOLD"``
        in :py:attr:`radis.config`  (``~/radis.json``)

        .. note::
            wstep = 'auto' is optimized for performances while ensuring accuracy,
            but is still experimental in 0.9.30. Feedback welcome!
    cutoff: float (~ unit of Linestrength: cm-1/(#.cm-2))
        discard linestrengths that are lower that this, to reduce calculation
        times. ``1e-27`` is what is generally used to generate databases such as
        CDSD. If ``0``, no cutoff. Default ``1e-27``.
    parsum_mode: 'full summation', 'tabulation'
        how to compute partition functions, at nonequilibrium or when partition
        function are not already tabulated. ``'full summation'`` : sums over all
        (potentially millions) of rovibrational levels. ``'tabulation'`` :
        builds an on-the-fly tabulation of rovibrational levels (500 - 4000x faster
        and usually accurate within 0.1%). Default ``full summation'``

        .. note::
            parsum_mode= 'tabulation'  is new in 0.9.30, and makes nonequilibrium
            calculations of small spectra extremelly fast. Will become the default
            after 0.9.31.
    pseudo_continuum_threshold: float
        if not ``0``, first calculate a rough approximation of the spectrum, then
        moves all lines whose linestrength intensity is less than this threshold
        of the maximum in a semi-continuum. Values above 0.01 can yield significant
        errors, mostly in highly populated areas. 80% of the lines can typically
        be moved in a continuum, resulting in 5 times faster spectra. If ``0``,
        no semi-continuum is used. Default ``0``.
    save_memory: boolean
        if ``True``, removes databases calculated by intermediate functions (for
        instance, delete the full database once the linestrength cutoff criteria
        was applied). This saves some memory but requires to reload the database
        & recalculate the linestrength for each new parameter. Default ``False``.
    export_populations: ``'vib'``, ``'rovib'``, ``None``
        if not None, store populations in Spectrum. Either store vibrational
        populations ('vib') or rovibrational populations ('rovib'). Default ``None``
    export_lines: boolean
        if ``True``, saves details of all calculated lines in Spectrum. This is
        necessary to later use :py:meth:`~radis.spectrum.spectrum.Spectrum.line_survey`,
        but can take some space. Default ``False``.
    chunksize: int, or ``None``
        Splits the lines database in several chuncks during calculation, else
        the multiplication of lines over all spectral range takes too much memory
        and slows the system down. Chunksize let you change the default chunck
        size. If ``None``, all lines are processed directly. Usually faster but
        can create memory problems. Default ``None``
    optimization : ``"simple"``, ``"min-RMS"``, ``None``
        If either ``"simple"`` or ``"min-RMS"`` DLM optimization for lineshape calculation is used:
        - ``"min-RMS"`` : weights optimized by analytical minimization of the RMS-error (See: [Spectral-Synthesis-Algorithm]_)
        - ``"simple"`` : weights equal to their relative position in the grid

        If using the DLM optimization, broadening method is automatically set to ``'fft'``.
        If ``None``, no lineshape interpolation is performed and the lineshape of all lines is calculated.

        Refer to [Spectral-Synthesis-Algorithm]_ for more explanation on the DLM method for lineshape interpolation.

        Default ``"min-RMS"``
    folding_thresh: float
        Folding is a correction procedure thet is applied when the lineshape is calculated with
        the ``fft`` broadening method and the linewidth is comparable to ``wstep``, that prevents
        sinc(v) modulation of the lineshape. Folding continues until the lineshape intensity
        is below ``folding_threshold``. Setting to 1 or higher effectively disables folding correction.

        Range: 0.0 < folding_thresh <= 1.0
        Default: 1e-6
    zero_padding: int
        Zero padding is used in conjunction with the ``fft`` broadening method to prevent circular
        convolution at the cost of performance. When set to -1, padding is set equal to the spectrum length,
        which guarantees a linear convolution.

        Range: 0 <= zero_padding <= len(w), or zero_padding = -1
        Default: -1
    broadening_method: ``"voigt"``, ``"convolve"``, ``"fft"``
        Calculates broadening with a direct voigt approximation ('voigt') or
        by convoluting independantly calculated Doppler and collisional
        broadening ('convolve'). First is much faster, 2nd can be used to
        compare results. This SpectrumFactory parameter can be manually
        adjusted a posteriori with::

            sf = SpectrumFactory(...)
            sf.params.broadening_method = 'voigt'

        Fast fourier transform ``'fft'`` is only available if using the DLM lineshape
        calculation ``optimization``. Because the DLM convolves all lines at the same time,
        and thus operates on large arrays, ``'fft'`` becomes more appropriate than
        convolutions in real space (``'voit'``, ``'convolve'`` )

        By default, use ``"fft"`` for any ``optimization``, and ``"voigt"`` if
        optimization is ``None`` .
    warnings: bool, or one of ``['warn', 'error', 'ignore']``, dict
        If one of ``['warn', 'error', 'ignore']``, set the default behaviour
        for all warnings. Can also be a dictionary to set specific warnings only.
        Example::

            warnings = {'MissingSelfBroadeningWarning':'ignore',
                        'NegativeEnergiesWarning':'ignore',
                        'HighTemperatureWarning':'ignore'}

        See :py:data:`~radis.misc.warning.default_warning_status` for more
        information.
    verbose: boolean, or int
        If ``False``, stays quiet. If ``True``, tells what is going on.
        If ``>=2``, gives more detailed messages (for instance, details of
        calculation times). Default ``True``.

    Examples
    --------

    An example using :class:`~radis.lbl.factory.SpectrumFactory`,
    :meth:`~radis.lbl.loader.DatabankLoader.load_databank`, the
    :class:`~radis.spectrum.spectrum.Spectrum` methods, and
    :py:mod:`~astropy.units` ::

        from radis import SpectrumFactory
        from astropy import units as u
        sf = SpectrumFactory(wavelength_min=4165 * u.nm,
                             wavelength_max=4200 * u.nm,
                             isotope='1,2',
                             truncation=10,  # cm-1
                             optimization=None,
                             medium='vacuum',
                             verbose=1,    # more for more details
                             )
        sf.load_databank('HITRAN-CO2-TEST')        # predefined in ~/radis.json
        s = sf.eq_spectrum(Tgas=300 * u.K, path_length=1 * u.cm)
        s.rescale_path_length(0.01)    # cm
        s.plot('radiance_noslit', Iunit='µW/cm2/sr/nm')

    Refer to the online :ref:`Examples <label_examples>` for more cases.

    .. minigallery:: radis.SpectrumFactory
        :add-heading:

    Notes
    -----

    .. inheritance-diagram:: radis.lbl.factory.SpectrumFactory
       :parts: 1

    High-level wrapper to SpectrumFactory:

    - :func:`~radis.lbl.calc.calc_spectrum`

    Main Methods:

    - :meth:`~radis.lbl.loader.DatabankLoader.load_databank`,
    - :meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum`,
    - :meth:`~radis.lbl.factory.SpectrumFactory.non_eq_spectrum`

    For advanced use:

    - :meth:`~radis.lbl.loader.DatabankLoader.fetch_databank`,
    - :meth:`~radis.lbl.loader.DatabankLoader.init_databank`,
    - :meth:`~radis.lbl.loader.DatabankLoader.init_database`,
    - :meth:`~radis.lbl.bands.BandFactory.eq_bands`,
    - :meth:`~radis.lbl.bands.BandFactory.non_eq_bands`

    Inputs and parameters can be accessed a posteriori with :

    - :py:attr:`~radis.lbl.loader.DatabankLoader.input` : physical input
    - :py:attr:`~radis.lbl.loader.DatabankLoader.params` : computational parameters
    - :py:attr:`~radis.lbl.loader.DatabankLoader.misc` : miscallenous parameters (don't change output)

    See Also
    --------
    :func:`~radis.lbl.calc.calc_spectrum`
    """

    # TODO: make it possible to export both 'vib' and 'rovib'

    # TODO
    # -------
    #
    # This code is still designed for one-molecule only: some parameters
    # (e.g.: molar mass used in Doppler broadening) are defined in SpectrumFactory
    # rather than being a column in the dataframe. This should be changed to
    # allow dealing with polymolecules databases

    # TODO
    # store everything in a self.var class instead of self.[] directly

    # TODO : move Tref in load_databank / fetch_databank only

    def __init__(
        self,
        wmin=None,
        wmax=None,
        wunit=Default("cm-1"),
        wavenum_min=None,
        wavenum_max=None,
        wavelength_min=None,
        wavelength_max=None,
        Tref=296,
        pressure=1.01325,
        mole_fraction=1,
        path_length=1,
        wstep=0.01,
        molecule=None,
        isotope="all",
        medium="air",
        truncation=Default(50),
        neighbour_lines=0,
        pseudo_continuum_threshold=0,
        self_absorption=True,
        chunksize=None,
        optimization="simple",
        folding_thresh=1e-6,
        zero_padding=-1,
        broadening_method=Default("voigt"),
        cutoff=0,
        parsum_mode="full summation",
        verbose=True,
        warnings=True,
        save_memory=False,
        export_populations=None,
        export_lines=False,
        **kwargs,
    ):

        # Initialize BandFactory object
        super(SpectrumFactory, self).__init__()

        # Check inputs (deal with deprecated format)
        if medium not in ["air", "vacuum"]:
            raise ValueError("Wavelength must be one of: 'air', 'vacuum'")
        kwargs0 = kwargs  # kwargs is used to deal with Deprecated names
        if "db_use_cached" in kwargs:
            warn(
                DeprecationWarning(
                    "db_use_cached removed from SpectrumFactory init and moved in load/fetch_databank()"
                )
            )
            kwargs0.pop("db_use_cached")
        if "lvl_use_cached" in kwargs:
            warn(
                DeprecationWarning(
                    "lvl_use_cached removed from SpectrumFactory init and moved in load/fetch_databank()"
                )
            )
            kwargs0.pop("lvl_use_cached")
        if "broadening_max_width" in kwargs:  # changed in 0.9.30
            broadening_max_width = kwargs["broadening_max_width"]
            raise (
                DeprecationWarning(
                    "`broadening_max_width`` (lineshape full-width, also used to compute the effect of neighbour lines) was replaced by `truncation` (lineshape half-width) and `neighbour_lines` (wavenumber range extension on each side). "
                    + f"To keep the current behavior, replace `broadening_max_width={broadening_max_width}` with "
                    + f"`truncation={broadening_max_width/2}, neighbour_lines={broadening_max_width/2}`. "
                    + "We recommended, for most cases: `truncation=300, neighbour_lines=0}`"
                )
            )

        if kwargs0 != {}:
            raise TypeError(
                "__init__() got an unexpected keyword argument '{0}'".format(
                    list(kwargs0)[0]
                )
            )

        if not 0 <= pseudo_continuum_threshold < 1:
            raise ValueError("pseudo_continuum_threshold should be in [0-1]")
        if export_populations not in ["vib", "rovib", False, None] and not isinstance(
            export_populations, list
        ):
            raise ValueError(
                "export_populations must be one of 'vib', 'rovib', "
                + "or 'False'. Got '{0}'".format(export_populations)
            )

        # calculate waveranges
        # --------------------

        # Get wavenumber, based on whatever was given as input.
        wavenum_min, wavenum_max = get_waverange(
            wmin,
            wmax,
            wunit,
            wavenum_min,
            wavenum_max,
            wavelength_min,
            wavelength_max,
            medium,
        )

        # Storing inital value of wstep if wstep != "auto"
        self.wstep = wstep

        # Init variables
        # --------------

        # Get molecule name
        if isinstance(molecule, int):
            molecule == get_molecule(molecule)
        if molecule is not None:
            if (
                molecule
                not in MOLECULES_LIST_EQUILIBRIUM + MOLECULES_LIST_NONEQUILIBRIUM
            ):
                raise ValueError(
                    "Unsupported molecule: {0}.\n".format(molecule)
                    + "Supported molecules are:\n - under equilibrium: {0}".format(
                        MOLECULES_LIST_EQUILIBRIUM
                    )
                    + "\n- under nonequilibrium: {0}".format(
                        MOLECULES_LIST_NONEQUILIBRIUM
                    )
                    + "\n\nNote that RADIS now has ExoMol support, but not all ExoMol molecules are referenced in RADIS. If a molecule is available in ExoMol but does not appear in RADIS yet, please contact the RADIS team or write on https://github.com/radis/radis/issues/319"
                )

        # Store isotope identifier in str format (list wont work in database queries)
        if not isinstance(isotope, str):
            isotope = ",".join([str(k) for k in list_if_float(isotope)])

        # Initialize input conditions
        self.input.wavenum_min = wavenum_min
        self.input.wavenum_max = wavenum_max
        self.input.Tref = convert_and_strip_units(Tref, u.K)
        self.input.pressure_mbar = convert_and_strip_units(pressure, u.bar) * 1e3
        self.input.mole_fraction = mole_fraction

        self.input.path_length = convert_and_strip_units(path_length, u.cm)
        self.input.molecule = (
            molecule  # if None, will be overwritten after reading database
        )
        self.input.state = "X"  # for the moment only ground-state is used
        # (but the code is electronic state aware)
        self.input.isotope = (
            isotope  # if 'all', will be overwritten after reading database
        )
        self.input.self_absorption = self_absorption

        # Initialize computation variables
        self.params.wstep = wstep
        self.params.pseudo_continuum_threshold = pseudo_continuum_threshold

        if cutoff is None:
            # If None, use no cutoff : https://github.com/radis/radis/pull/259
            cutoff = 0
        self.params.cutoff = cutoff
        self.params.parsum_mode = parsum_mode

        # Time Based variables
        self.verbose = verbose

        # if optimization is ``'simple'`` or ``'min-RMS'``, or None :
        # Adjust default values of broadening method :
        if isinstance(broadening_method, Default):
            if optimization in ("simple", "min-RMS") and broadening_method != "voigt":
                if self.verbose >= 3:
                    printg(
                        "LDM algorithm used. Defaulting broadening method from {0} to Voigt".format(
                            broadening_method
                        )
                    )
                broadening_method = "voigt"
            elif optimization is None and broadening_method != "voigt":
                if self.verbose >= 3:
                    printg(
                        "LDM algorithm not used. Defaulting broadening method from {0} to 'voigt'".format(
                            broadening_method
                        )
                    )
                broadening_method = "voigt"
            else:  # keep default
                broadening_method = broadening_method.value

        if truncation == 0:
            raise ValueError(
                "Lineshape truncation must be >0. If you want no truncation (compute lineshape on the full spectral range), use `truncation=None`"
            )

        self.misc.export_lines = export_lines
        self.misc.export_populations = export_populations

        if broadening_method == "fft":
            if isinstance(truncation, Default):
                truncation = None
            else:
                if broadening_method == "fft":
                    raise NotImplementedError(
                        "Lines cannot be truncated with `broadening_method='fft'`. Use `broadening_method='voigt'`"
                    )
        if isinstance(truncation, Default):
            truncation = truncation.value

        self.params.truncation = self.truncation = truncation  # line truncation
        # self.params.truncation is the input, self.truncation will be the value (different from input if input was None)
        self.params.neighbour_lines = neighbour_lines  # including neighbour lines

        # # reduce neighbour_lines if unnecessary
        # if truncation and truncation < neighbour_lines:
        #     self.warn
        # Define max range on which to load lines (required for fetch_databank / load_databank)
        self.params.wavenum_min_calc = wavenum_min - neighbour_lines
        self.params.wavenum_max_calc = wavenum_max + neighbour_lines
        #  note @dev : if neighbour_lines changes during the SpectrumFactory loop (e.g:  user sets sf.params.neighbour_lines= ;  the database wont be updated.
        # (databank should be reloaded). We store and check later to ensure it is not changed afterwards
        self._neighbour_lines = neighbour_lines

        self.params.broadening_method = broadening_method
        self.params.optimization = optimization
        self.params.folding_thresh = folding_thresh
        self.misc.zero_padding = zero_padding

        # used to split lines into blocks not too big for memory
        self.misc.chunksize = chunksize
        # Other parameters:
        self.save_memory = save_memory
        self.autoupdatedatabase = False  # a boolean to automatically store calculated
        # spectra in a Spectrum database. See init_database
        # for more info
        self.autoretrievedatabase = False  # a boolean to automatically retrieve
        # spectra from database instead of
        # calculating them
        self.SpecDatabase = None  # the database to store spectra. Not to be confused
        # with the databank where lines are stored
        self.database = None  # path to previous database
        # Warnings
        # --------

        # Add cutoff / threshold warning values for the different optimisations
        # (linestrength cutoff, broadening cutoff). These cannot be changed
        # from the Factory input, but can still be modified manually afterwards
        # TODO: replace everything with 'auto' modes.

        # Set default behavior for warnings:
        if isinstance(warnings, dict):
            self.warnings.update(warnings)
        elif warnings in [True, "warn", "warning"]:
            self.warnings["default"] = "warn"
        elif warnings == "error":
            self.warnings["default"] = "error"
        elif warnings in [False, "ignore"]:
            self.warnings["default"] = "ignore"
        else:
            raise ValueError("Unexpected value for warnings: {0}".format(warnings))
        # Set default values for warnings thresholds
        self.misc.warning_linestrength_cutoff = 1e-2
        self.misc.warning_broadening_threshold = 1e-2

        return

    # %% ======================================================================
    # PUBLIC METHODS
    # ------------------------
    # eq_spectrum             >>> calc equilibrium spectrum
    # non_eq_spectrum         >>> calc non equilibrium spectrum
    # eq_bands                >>> returns bands as a list of spectra
    # non_eq_bands            >>> same, with overpopulation factors
    # power                   >>> get total power (equilibrium or non eq)
    # fit_spectrum            >>> fit experimental spectrum
    #
    # XXX =====================================================================

    def eq_spectrum(
        self, Tgas, mole_fraction=None, path_length=None, pressure=None, name=None
    ) -> Spectrum:
        """Generate a spectrum at equilibrium.

        Parameters
        ----------
        Tgas: float or `~astropy.units.quantity.Quantity`
            Gas temperature (K)
        mole_fraction: float
            database species mole fraction. If None, Factory mole fraction is used.
        path_length: float or `~astropy.units.quantity.Quantity`
            slab size (cm). If ``None``, the default Factory
            :py:attr:`~radis.lbl.factory.SpectrumFactor.input.path_length` is used.
        pressure: float or `~astropy.units.quantity.Quantity`
            pressure (bar). If ``None``, the default Factory
            :py:attr:`~radis.lbl.factory.SpectrumFactor.input.pressure` is used.
        name: str
            output Spectrum name (useful in batch)

        Returns
        -------
        s : Spectrum
            Returns a :class:`~radis.spectrum.spectrum.Spectrum` object

                Use the :meth:`~radis.spectrum.spectrum.Spectrum.get` method to get something
                among ``['radiance', 'radiance_noslit', 'absorbance', etc...]``

                Or directly the :meth:`~radis.spectrum.spectrum.Spectrum.plot` method
                to plot it. See [1]_ to get an overview of all Spectrum methods

        Examples
        --------
        ::
            from radis import SpectrumFactory
            sf = SpectrumFactory(
            wavenum_min=2900,
            wavenum_max=3200,
            molecule="OH",
            wstep=0.1,
            )
            sf.fetch_databank("hitemp")

            s1 = sf.eq_spectrum(Tgas=300, path_length=1, pressure=0.1)
            s2 = sf.eq_spectrum(Tgas=500, path_length=1, pressure=0.1)

        .. minigallery:: radis.lbl.SpectrumFactory.eq_spectrum
            :add-heading:

        References
        ----------
        .. [1] RADIS doc: `Spectrum how to? <https://radis.readthedocs.io/en/latest/spectrum/spectrum.html#label-spectrum>`__


        See Also
        --------
        :meth:`~radis.lbl.factory.SpectrumFactory.non_eq_spectrum`
        """

        # %% Preprocessing
        # --------------------------------------------------------------------

        # Check inputs
        if not self.input.self_absorption:
            raise ValueError(
                "Use non_eq_spectrum(Tgas, Tgas) to calculate spectra "
                + "without self_absorption"
            )

        # Convert units
        Tgas = convert_and_strip_units(Tgas, u.K)
        path_length = convert_and_strip_units(path_length, u.cm)
        pressure = convert_and_strip_units(pressure, u.bar)

        # update defaults
        if path_length is not None:
            self.input.path_length = path_length
        if mole_fraction is not None:
            self.input.mole_fraction = mole_fraction
        if pressure is not None:
            self.input.pressure_mbar = pressure * 1e3
        if not is_float(Tgas):
            raise ValueError(
                "Tgas should be float or Astropy unit. Got {0}".format(Tgas)
            )
        self.input.rot_distribution = "boltzmann"  # equilibrium
        self.input.vib_distribution = "boltzmann"  # equilibrium

        # Get temperatures
        self.input.Tgas = Tgas
        self.input.Tvib = Tgas  # just for info
        self.input.Trot = Tgas  # just for info

        # Init variables
        pressure_mbar = self.input.pressure_mbar
        mole_fraction = self.input.mole_fraction
        path_length = self.input.path_length
        verbose = self.verbose

        # New Profiler object
        self._reset_profiler(verbose)

        # Check variables
        self._check_inputs(mole_fraction, max(flatten(Tgas)))

        # Retrieve Spectrum from database if it exists
        if self.autoretrievedatabase:
            s = self._retrieve_from_database()
            if s is not None:
                return s  # exit function

        # %% Start
        # --------------------------------------------------------------------

        self.profiler.start("spectrum_calculation", 1)
        self.profiler.start("spectrum_calc_before_obj", 2)

        if verbose:
            self.print_conditions("Calculating Equilibrium Spectrum")

        # Check database, reset populations, create line dataframe to be scaled
        # --------------------------------------------------------------------
        self._check_line_databank()
        self._reinitialize()  # creates scaled dataframe df1 from df0

        # --------------------------------------------------------------------

        # First calculate the linestrength at given temperature
        self.calc_linestrength_eq(Tgas)  # scales S0 to S (equivalent to S0 in code)
        self._cutoff_linestrength()

        # ----------------------------------------------------------------------

        # Calculate line shift
        self.calc_lineshift()  # scales wav to shiftwav (equivalent to v0)

        # ----------------------------------------------------------------------
        # Line broadening

        # ... calculate broadening  HWHM
        self._calc_broadening_HWHM()

        # ... generates all wstep related entities
        self._generate_wavenumber_arrays()

        # ... find weak lines and calculate semi-continuum (optional)
        I_continuum = self.calculate_pseudo_continuum()
        # ... apply lineshape and get absorption coefficient
        # ... (this is the performance bottleneck)
        wavenumber, abscoeff_v = self._calc_broadening()
        #    :         :
        #   cm-1    1/(#.cm-2)

        # ... add semi-continuum (optional)
        abscoeff_v = self._add_pseudo_continuum(abscoeff_v, I_continuum)
        # Calculate output quantities
        # ----------------------------------------------------------------------

        self.profiler.start("calc_other_spectral_quan", 2)

        # incorporate density of molecules (see equation (A.16) )
        density = mole_fraction * ((pressure_mbar * 100) / (k_b * Tgas)) * 1e-6
        #  :
        # (#/cm3)

        abscoeff = abscoeff_v * density  # cm-1
        # ... # TODO: if the code is extended to multi-species, then density has to be added
        # ... before lineshape broadening (as it would not be constant for all species)

        # get absorbance (technically it's the optical depth `tau`,
        #                absorbance `A` being `A = tau/ln(10)` )
        absorbance = abscoeff * path_length
        # Generate output quantities
        transmittance_noslit = exp(-absorbance)
        emissivity_noslit = 1 - transmittance_noslit
        radiance_noslit = calc_radiance(
            wavenumber, emissivity_noslit, Tgas, unit=self.units["radiance_noslit"]
        )

        self.profiler.stop(
            "calc_other_spectral_quan", "Calculated other spectral quantities"
        )

        # %% Export
        # --------------------------------------------------------------------

        self.profiler.stop(
            "spectrum_calc_before_obj", "Spectrum calculated (before object generation)"
        )
        self.profiler.start("generate_spectrum_obj", 2)

        # Get conditions
        conditions = self.get_conditions()
        conditions.update(
            {
                "calculation_time": self.profiler.final[list(self.profiler.final)[-1]][
                    "spectrum_calc_before_obj"
                ],
                "lines_calculated": self._Nlines_calculated,
                "lines_cutoff": self._Nlines_cutoff,
                "lines_in_continuum": self._Nlines_in_continuum,
                "thermal_equilibrium": True,
                "radis_version": version,
                "spectral_points": (
                    self.params.wavenum_max_calc - self.params.wavenum_min_calc
                )
                / self.params.wstep,
                "profiler": dict(self.profiler.final),
            }
        )
        if self.params.optimization != None:
            conditions.update(
                {
                    "wL": self.wL,
                    "wG": self.wG,
                }
            )
        del self.profiler.final[list(self.profiler.final)[-1]][
            "spectrum_calc_before_obj"
        ]

        # Get populations of levels as calculated in RovibrationalPartitionFunctions
        # ... Populations cannot be calculated at equilibrium (needs energies).
        # ... Use SpectrumFactory.non_eq_spectrum
        populations = None

        # Get lines (intensities + populations)
        lines = self.get_lines()

        # Spectral quantities
        quantities = {
            "abscoeff": (wavenumber, abscoeff),
            "absorbance": (wavenumber, absorbance),
            "emissivity_noslit": (wavenumber, emissivity_noslit),
            "transmittance_noslit": (wavenumber, transmittance_noslit),
            # (mW/cm2/sr/nm)
            "radiance_noslit": (wavenumber, radiance_noslit),
        }
        if I_continuum is not None and self._export_continuum:
            quantities.update(
                {"abscoeff_continuum": (wavenumber, I_continuum * density)}
            )

        # Store results in Spectrum class
        s = Spectrum(
            quantities=quantities,
            conditions=conditions,
            populations=populations,
            lines=lines,
            units=self.units,
            cond_units=self.cond_units,
            wunit=self.params.waveunit,  # cm-1
            # dont check input (much faster, and Spectrum
            check_wavespace=False,
            # is freshly baken so probably in a good format
            name=name,
            references=dict(self.reftracker),
        )

        # update database if asked so
        if self.autoupdatedatabase:
            self.SpecDatabase.add(s, if_exists_then="increment")
            # Tvib=Trot=Tgas... but this way names in a database
            # generated with eq_spectrum are consistent with names
            # in one generated with non_eq_spectrum

        # Get generation & total calculation time
        self.profiler.stop("generate_spectrum_obj", "Generated Spectrum object")

        #  In the less verbose case, we print the total calculation+generation time:
        self.profiler.stop("spectrum_calculation", "Spectrum calculated")

        return s

    def eq_spectrum_gpu(
        self, Tgas, mole_fraction=None, path_length=None, pressure=None, name=None
    ) -> Spectrum:
        """Generate a spectrum at equilibrium with calculation of lineshapes
        and broadening done on the GPU.

        .. note::
            This method requires CUDA compatible hardware to execute.
            For more information on how to setup your system to run GPU-accelerated methods
            using CUDA and Cython, check `GPU Spectrum Calculation on RADIS <https://radis.readthedocs.io/en/latest/lbl/gpu.html>`

        Parameters
        ----------
        Tgas: float or `~astropy.units.quantity.Quantity`
            Gas temperature (K)
        mole_fraction: float
            database species mole fraction. If None, Factory mole fraction is used.
        path_length: float or `~astropy.units.quantity.Quantity`
            slab size (cm). If ``None``, the default Factory
            :py:attr:`~radis.lbl.factory.SpectrumFactor.input.path_length` is used.
        pressure: float or `~astropy.units.quantity.Quantity`
            pressure (bar). If ``None``, the default Factory
            :py:attr:`~radis.lbl.factory.SpectrumFactor.input.pressure` is used.
        name: str
            output Spectrum name (useful in batch)

        Returns
        -------
        s : Spectrum
            Returns a :class:`~radis.spectrum.spectrum.Spectrum` object

                Use the :meth:`~radis.spectrum.spectrum.Spectrum.get` method to get something
                among ``['radiance', 'radiance_noslit', 'absorbance', etc...]``

                Or directly the :meth:`~radis.spectrum.spectrum.Spectrum.plot` method
                to plot it. See [1]_ to get an overview of all Spectrum methods

        See Also
        --------
        :meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum`
        """

        # %% Preprocessing
        # --------------------------------------------------------------------

        # Check inputs
        if not self.input.self_absorption:
            raise ValueError(
                "Use non_eq_spectrum(Tgas, Tgas) to calculate spectra "
                + "without self_absorption"
            )

        # Convert units
        Tgas = convert_and_strip_units(Tgas, u.K)
        path_length = convert_and_strip_units(path_length, u.cm)
        pressure = convert_and_strip_units(pressure, u.bar)

        # update defaults
        if path_length is not None:
            self.input.path_length = path_length
        if mole_fraction is not None:
            self.input.mole_fraction = mole_fraction
        if pressure is not None:
            self.input.pressure_mbar = pressure * 1e3
        if not is_float(Tgas):
            raise ValueError("Tgas should be float.")
        self.input.rot_distribution = "boltzmann"  # equilibrium
        self.input.vib_distribution = "boltzmann"  # equilibrium

        # Get temperatures
        self.input.Tgas = Tgas
        self.input.Tvib = Tgas  # just for info
        self.input.Trot = Tgas  # just for info

        verbose = self.verbose

        # New Profiler object
        self._reset_profiler(verbose)

        # Init variables
        pressure_mbar = self.input.pressure_mbar
        mole_fraction = self.input.mole_fraction
        path_length = self.input.path_length

        # Check variables
        self._check_inputs(mole_fraction, max(flatten(Tgas)))

        # Retrieve Spectrum from database if it exists
        if self.autoretrievedatabase:
            s = self._retrieve_from_database()
            if s is not None:
                return s  # exit function

        ### GET ISOTOPE ABUNDANCE & MOLECULAR MASS ###

        molpar = self.molparam

        try:
            if "id" in self.df0:
                id_set = self.df0[
                    "id"
                ].unique()  # get all the molecules in the dataframe, should ideally be 1 element for GPU
                mol_id = id_set[0]

                assert len(id_set) == 1  # make sure there is only one molecule

            else:
                mol_id = self.df0.attrs["id"]
            molecule = get_molecule(mol_id)
        except:
            mol_id = get_molecule_identifier(self.input.molecule)
            molecule = get_molecule(mol_id)

        state = self.input.state
        iso_set = self._get_isotope_list(molecule)

        iso_arr = list(range(max(iso_set) + 1))

        Ia_arr = np.empty_like(iso_arr, dtype=np.float32)  # abundance of each isotope
        molarmass_arr = np.empty_like(
            iso_arr, dtype=np.float32
        )  # molar mass of each isotope
        Q_arr = np.empty_like(
            iso_arr, dtype=np.float32
        )  # partitioning function of each isotope
        for iso in iso_arr:
            if iso in iso_set:
                params = molpar.df.loc[(mol_id, iso)]
                Ia_arr[iso] = params.abundance
                molarmass_arr[iso] = params.mol_mass
                Q_arr[iso] = self._get_parsum(molecule, iso, state).at(T=Tgas)

        Ia_arr[np.isnan(Ia_arr)] = 0
        molarmass_arr[np.isnan(molarmass_arr)] = 0

        self.profiler.start("spectrum_calculation", 1)
        self.profiler.start("spectrum_calc_before_obj", 2)

        # generate the v_arr
        v_arr = np.arange(
            self.input.wavenum_min,
            self.input.wavenum_max + self.params.wstep,
            self.params.wstep,
        )

        # load the data
        df = self.df0
        iso = df["iso"].to_numpy(dtype=np.int32)
        v0 = df["wav"].to_numpy(dtype=np.float32)
        da = df["Pshft"].to_numpy(dtype=np.float32)
        El = df["El"].to_numpy(dtype=np.float32)
        na = df["Tdpair"].to_numpy(dtype=np.float32)

        log_2gs = np.array(self._get_log_2gs(), dtype=np.float32)
        log_2vMm = np.array(self._get_log_2vMm(molarmass_arr), dtype=np.float32)
        S0 = np.array(self._get_S0(Ia_arr), dtype=np.float32)

        NwG = 4  # TO-DO: these shouldn't be hardcoded
        NwL = 8  # TO-DO: these shouldn't be hardcoded

        _Nlines_calculated = len(v0)

        if verbose >= 2:
            print("Initializing parameters...", end=" ")

        if verbose is False:
            verbose_gpu = 0
        elif verbose is True:
            verbose_gpu = 1
        else:
            verbose_gpu = verbose

        try:
            from radis.lbl.gpu import gpu_init, gpu_iterate
        except (ModuleNotFoundError):
            print("Failed to load GPU module, exiting!")
            exit()

        gpu_init(
            v_arr,
            NwG,
            NwL,
            iso,
            v0,
            da,
            log_2gs,
            na,
            log_2vMm,
            S0,
            El,
            Q_arr,
            verbose_gpu,
        )

        if verbose >= 2:
            print("Initialization complete!")

        wavenumber = v_arr

        if verbose >= 2:
            print("Calculating spectra...", end=" ")

        abscoeff = gpu_iterate(
            pressure_mbar * 1e-3,
            Tgas,
            mole_fraction,
            Ia_arr,
            molarmass_arr,
            verbose_gpu,
        )
        # Calculate output quantities
        # ----------------------------------------------------------------------

        self.profiler.start("calc_other_spectral_quan", 2)

        # ... # TODO: if the code is extended to multi-species, then density has to be added
        # ... before lineshape broadening (as it would not be constant for all species)

        # get absorbance (technically it's the optical depth `tau`,
        #                absorbance `A` being `A = tau/ln(10)` )
        absorbance = abscoeff * path_length
        # Generate output quantities
        transmittance_noslit = exp(-absorbance)
        emissivity_noslit = 1 - transmittance_noslit
        radiance_noslit = calc_radiance(
            wavenumber, emissivity_noslit, Tgas, unit=self.units["radiance_noslit"]
        )

        self.profiler.stop(
            "calc_other_spectral_quan", "Calculated other spectral quantities"
        )

        lines = self.get_lines()

        # %% Export
        # --------------------------------------------------------------------

        self.profiler.stop(
            "spectrum_calc_before_obj", "Spectrum calculated (before object generation)"
        )
        self.profiler.start("generate_spectrum_obj", 2)

        # Get lines (intensities + populations)

        conditions = self.get_conditions()
        conditions.update(
            {
                "calculation_time": self.profiler.final[list(self.profiler.final)[-1]][
                    "spectrum_calc_before_obj"
                ],
                "lines_calculated": _Nlines_calculated,
                "thermal_equilibrium": True,
                "radis_version": version,
                "spectral_points": (
                    self.params.wavenum_max_calc - self.params.wavenum_min_calc
                )
                / self.params.wstep,
                "profiler": dict(self.profiler.final),
            }
        )
        if self.params.optimization != None:
            conditions.update(
                {
                    "wL": self.wL,
                    "wG": self.wG,
                }
            )
        del self.profiler.final[list(self.profiler.final)[-1]][
            "spectrum_calc_before_obj"
        ]

        # Spectral quantities
        quantities = {
            "abscoeff": (wavenumber, abscoeff),
            "absorbance": (wavenumber, absorbance),
            "emissivity_noslit": (wavenumber, emissivity_noslit),
            "transmittance_noslit": (wavenumber, transmittance_noslit),
            # (mW/cm2/sr/nm)
            "radiance_noslit": (wavenumber, radiance_noslit),
        }

        # Store results in Spectrum class
        s = Spectrum(
            quantities=quantities,
            units=self.units,
            conditions=conditions,
            lines=lines,
            cond_units=self.cond_units,
            wunit=self.params.waveunit,  # cm-1
            # dont check input (much faster, and Spectrum
            check_wavespace=False,
            # is freshly baken so probably in a good format
            name=name,
            references=dict(self.reftracker),
        )

        # update database if asked so
        if self.autoupdatedatabase:
            self.SpecDatabase.add(s, if_exists_then="increment")
            # Tvib=Trot=Tgas... but this way names in a database
            # generated with eq_spectrum are consistent with names
            # in one generated with non_eq_spectrum

        # Get generation & total calculation time
        self.profiler.stop("generate_spectrum_obj", "Generated Spectrum object")

        #  In the less verbose case, we print the total calculation+generation time:
        self.profiler.stop("spectrum_calculation", "Spectrum calculated")

        return s

    def non_eq_spectrum(
        self,
        Tvib,
        Trot,
        Ttrans=None,
        mole_fraction=None,
        path_length=None,
        pressure=None,
        vib_distribution="boltzmann",
        rot_distribution="boltzmann",
        overpopulation=None,
        name=None,
    ) -> Spectrum:
        """Calculate emission spectrum in non-equilibrium case. Calculates
        absorption with broadened linestrength and emission with broadened
        Einstein coefficient.

        Parameters
        ----------
        Tvib: float
            vibrational temperature [K]
            can be a tuple of float for the special case of more-than-diatomic
            molecules (e.g: CO2)
        Trot: float
            rotational temperature [K]
        Ttrans: float
            translational temperature [K]. If None, translational temperature is
            taken as rotational temperature (valid at 1 atm for times above ~ 2ns
            which is the RT characteristic time)
        mole_fraction: float
            database species mole fraction. If None, Factory mole fraction is used.
        path_length: float or `~astropy.units.quantity.Quantity`
            slab size (cm). If ``None``, the default Factory
            :py:attr:`~radis.lbl.factory.SpectrumFactor.input.path_length` is used.
        pressure: float or `~astropy.units.quantity.Quantity`
            pressure (bar). If ``None``, the default Factory
            :py:attr:`~radis.lbl.factory.SpectrumFactor.input.pressure` is used.

        Other Parameters
        ----------------
        vib_distribution: ``'boltzmann'``, ``'treanor'``
            vibrational distribution
        rot_distribution: ``'boltzmann'``
            rotational distribution
        overpopulation: dict, or ``None``
            add overpopulation factors for given levels::

                {level:overpopulation_factor}
        name: str
            output Spectrum name (useful in batch)

        Returns
        -------
        s : Spectrum
            Returns a :class:`~radis.spectrum.spectrum.Spectrum` object

                Use the :meth:`~radis.spectrum.spectrum.Spectrum.get` method to get something
                among ``['radiance', 'radiance_noslit', 'absorbance', etc...]``

                Or directly the :meth:`~radis.spectrum.spectrum.Spectrum.plot` method
                to plot it. See [1]_ to get an overview of all Spectrum methods

        Examples
        --------
        ::
            from radis import SpectrumFactory
            sf = SpectrumFactory(
            wavenum_min=2000,
            wavenum_max=3000,
            molecule="CO",
            wstep=0.1,
            )
            sf.fetch_databank("hitemp")

            s1 = sf.non_eq_spectrum(Tvib=2000, Trot=600, path_length=1, pressure=0.1)
            s2 = sf.non_eq_spectrum(Tvib=2000, Trot=600, path_length=1, pressure=0.1)

        .. minigallery:: radis.lbl.SpectrumFactory.non_eq_spectrum
            :add-heading:

        References
        ----------
        .. [1] RADIS doc: `Spectrum how to? <https://radis.readthedocs.io/en/latest/spectrum/spectrum.html#label-spectrum>`__

        See Also
        --------

        :meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum`
        :meth:`~radis.lbl.factory.SpectrumFactory.optically_thin_power`
        """

        # %% Preprocessing
        # --------------------------------------------------------------------

        # Convert units
        Tvib = convert_and_strip_units(Tvib, u.K)
        Trot = convert_and_strip_units(Trot, u.K)
        Ttrans = convert_and_strip_units(Ttrans, u.K)
        path_length = convert_and_strip_units(path_length, u.cm)
        pressure = convert_and_strip_units(pressure, u.bar)

        # check inputs, update defaults
        if path_length is not None:
            self.input.path_length = path_length
        if mole_fraction is not None:
            self.input.mole_fraction = mole_fraction
        if pressure is not None:
            self.input.pressure_mbar = pressure * 1e3
        if isinstance(Tvib, tuple):
            Tvib = tuple([convert_and_strip_units(T, u.K) for T in Tvib])
        elif not is_float(Tvib):
            raise TypeError(
                "Tvib should be float, or tuple (got {0})".format(type(Tvib))
            )
        singleTvibmode = is_float(Tvib)
        if not is_float(Trot):
            raise ValueError("Trot should be float")
        if overpopulation is None:
            overpopulation = {}
        assert vib_distribution in ["boltzmann", "treanor"]
        assert rot_distribution in ["boltzmann"]
        self.input.overpopulation = overpopulation
        self.input.rot_distribution = rot_distribution
        self.input.vib_distribution = vib_distribution

        # Get translational temperature
        Tgas = Ttrans
        if Tgas is None:
            Tgas = Trot  # assuming Ttrans = Trot
        self.input.Tgas = Tgas
        self.input.Tvib = Tvib
        self.input.Trot = Trot

        # Init variables
        path_length = self.input.path_length
        mole_fraction = self.input.mole_fraction
        pressure_mbar = self.input.pressure_mbar
        verbose = self.verbose

        # New Profiler object
        self._reset_profiler(verbose)

        # Check variables
        self._check_inputs(mole_fraction, max(flatten(Tgas, Tvib, Trot)))

        # Retrieve Spectrum from database if it exists
        if self.autoretrievedatabase:
            s = self._retrieve_from_database()
            if s is not None:
                return s  # exit function

        # %% Start
        # --------------------------------------------------------------------

        self.profiler.start("spectrum_calculation", 1)
        self.profiler.start("spectrum_calc_before_obj", 2)
        if verbose:
            self.print_conditions("Calculating Non-Equilibrium Spectrum")

        # Check line database and parameters, reset populations and scaled line dataframe
        # ----------
        self._check_line_databank()

        # add nonequilibrium energies if needed (this may be a bottleneck
        # for a first calculation):
        self._calc_noneq_parameters(vib_distribution, singleTvibmode)
        self._reinitialize()  # creates scaled dataframe df1 from df0

        # ----------------------------------------------------------------------
        # Calculate Populations, Linestrength and Emission Integral
        if singleTvibmode:
            self.calc_populations_noneq(
                Tvib,
                Trot,
                vib_distribution=vib_distribution,
                rot_distribution=rot_distribution,
                overpopulation=overpopulation,
            )
        else:
            self._calc_populations_noneq_multiTvib(
                Tvib,
                Trot,
                vib_distribution=vib_distribution,
                rot_distribution=rot_distribution,
                overpopulation=overpopulation,
            )

        self.calc_linestrength_noneq()
        self.calc_emission_integral()

        # ----------------------------------------------------------------------
        # Cutoff linestrength
        self._cutoff_linestrength()

        # ----------------------------------------------------------------------

        # Calculate lineshift
        self.calc_lineshift()

        # ----------------------------------------------------------------------

        # Line broadening

        # ... calculate broadening  HWHM
        self._calc_broadening_HWHM()

        # ... generates all wstep related entities
        self._generate_wavenumber_arrays()

        # ... find weak lines and calculate semi-continuum (optional)
        k_continuum, j_continuum = self.calculate_pseudo_continuum(noneq=True)

        # ... apply lineshape and get absorption coefficient
        # ... (this is the performance bottleneck)
        wavenumber, abscoeff_v, emisscoeff_v = self._calc_broadening_noneq()
        #    :         :            :
        #   cm-1    1/(#.cm-2)   mW/sr/cm-1

        # ... add semi-continuum (optional)
        abscoeff_v = self._add_pseudo_continuum(abscoeff_v, k_continuum)
        emisscoeff_v = self._add_pseudo_continuum(emisscoeff_v, j_continuum)

        # Calculate output quantities
        # ----------------------------------------------------------------------

        self.profiler.start("calc_other_spectral_quan", 2)

        # incorporate density of molecules (see Rothman 1996 equation (A.16) )
        density = mole_fraction * ((pressure_mbar * 100) / (k_b * Tgas)) * 1e-6
        #  :
        # (#/cm3)

        abscoeff = abscoeff_v * density  # cm-1
        emisscoeff = emisscoeff_v * density  # mW/sr/cm3/cm-1

        # ... # TODO: if the code is extended to multi-species, then density has to be added
        # ... before lineshape broadening (as it would not be constant for all species)

        # get absorbance (technically it's the optical depth `tau`,
        #                absorbance `A` being `A = tau/ln(10)` )

        # Generate output quantities
        absorbance = abscoeff * path_length  # (adim)
        transmittance_noslit = exp(-absorbance)

        if self.input.self_absorption:
            # Analytical output of computing RTE over a single slab of constant
            # emissivity and absorption coefficient
            b = abscoeff == 0  # optically thin mask
            radiance_noslit = np.zeros_like(emisscoeff)
            radiance_noslit[~b] = (
                emisscoeff[~b] / abscoeff[~b] * (1 - transmittance_noslit[~b])
            )
            radiance_noslit[b] = emisscoeff[b] * path_length
        else:
            # Note that for k -> 0,
            radiance_noslit = emisscoeff * path_length  # (mW/sr/cm2/cm-1)

        # Convert `radiance_noslit` from (mW/sr/cm2/cm-1) to (mW/sr/cm2/nm)
        radiance_noslit = convert_rad2nm(
            radiance_noslit, wavenumber, "mW/sr/cm2/cm-1", "mW/sr/cm2/nm"
        )
        # Convert 'emisscoeff' from (mW/sr/cm3/cm-1) to (mW/sr/cm3/nm)
        emisscoeff = convert_emi2nm(
            emisscoeff, wavenumber, "mW/sr/cm3/cm-1", "mW/sr/cm3/nm"
        )

        self.profiler.stop(
            "calc_other_spectral_quan", "Calculated other spectral quantities"
        )

        # Note: emissivity not defined under non equilibrium

        # %% Export
        # ----------------------------------------------------------------------

        self.profiler.stop(
            "spectrum_calc_before_obj", "Spectrum calculated (before object generation)"
        )
        self.profiler.start("generate_spectrum_obj", 2)

        # Get conditions
        conditions = self.get_conditions()
        conditions.update(
            {
                "calculation_time": self.profiler.final[list(self.profiler.final)[-1]][
                    "spectrum_calc_before_obj"
                ],
                "lines_calculated": self._Nlines_calculated,
                "lines_cutoff": self._Nlines_cutoff,
                "lines_in_continuum": self._Nlines_in_continuum,
                "thermal_equilibrium": False,  # dont even try to guess if it's at equilibrium
                "radis_version": version,
                "spectral_points": (
                    self.params.wavenum_max_calc - self.params.wavenum_min_calc
                )
                / self.params.wstep,
                "profiler": dict(self.profiler.final),
            }
        )
        if self.params.optimization != None:
            conditions.update(
                {
                    "wL": self.wL,
                    "wG": self.wG,
                }
            )
        del self.profiler.final[list(self.profiler.final)[-1]][
            "spectrum_calc_before_obj"
        ]

        # Get populations of levels as calculated in RovibrationalPartitionFunctions
        populations = self.get_populations(self.misc.export_populations)

        # Get lines (intensities + populations)
        lines = self.get_lines()

        # Spectral quantities
        quantities = {
            "abscoeff": (wavenumber, abscoeff),
            "absorbance": (wavenumber, absorbance),
            # (mW/cm3/sr/nm)
            "emisscoeff": (wavenumber, emisscoeff),
            "transmittance_noslit": (wavenumber, transmittance_noslit),
            # (mW/cm2/sr/nm)
            "radiance_noslit": (wavenumber, radiance_noslit),
        }
        if k_continuum is not None and self._export_continuum:
            quantities.update(
                {
                    "abscoeff_continuum": (wavenumber, k_continuum * density),
                    "emisscoeff_continuum": (wavenumber, j_continuum * density),
                }
            )

        # Store results in Spectrum class
        s = Spectrum(
            quantities=quantities,
            conditions=conditions,
            populations=populations,
            lines=lines,
            units=self.units,
            cond_units=self.cond_units,
            wunit=self.params.waveunit,  # cm-1
            # dont check input (much faster, and Spectrum
            check_wavespace=False,
            # is freshly baken so probably in a good format
            name=name,
            references=dict(self.reftracker),
        )

        # update database if asked so
        if self.autoupdatedatabase:
            self.SpecDatabase.add(
                s,
                add_info=["Tvib", "Trot"],
                add_date="%Y%m%d",
                if_exists_then="increment",
            )

        # Get generation & total calculation time
        self.profiler.stop("generate_spectrum_obj", "Generated Spectrum object")

        #  In the less verbose case, we print the total calculation+generation time:
        self.profiler.stop("spectrum_calculation", "Spectrum calculated")

        return s

    def _generate_wavenumber_arrays(self):
        """define wavenumber grid vectors

        `SpectrumFactory.wavenumber` is the output spectral range and
        ``SpectrumFactory.wavenumber_calc`` the spectral range used for calculation, that
        includes neighbour lines within ``neighbour_lines`` distance."""

        self.profiler.start("generate_wavenumber_arrays", 2)
        # calculates minimum FWHM of lines
        self._calc_min_width(self.df1)

        # Setting wstep to optimal value and rounding it to a degree 3
        if self.wstep == "auto":
            import radis

            self.params.wstep = round_off(
                self.min_width / radis.config["GRIDPOINTS_PER_LINEWIDTH_WARN_THRESHOLD"]
            )
            self.warnings["AccuracyWarning"] = "ignore"

        truncation = self.params.truncation
        neighbour_lines = self.params.neighbour_lines

        if truncation and neighbour_lines > truncation:
            self.warn(
                f"Neighbour lines resolved up to {neighbour_lines} cm-1 away from the spectrum. "
                + f"But lines are anyway truncated at {truncation:.2f} cm-1. "
                + f"Choose 'neighbour_lines={truncation:.2f}' to avoid resolving useless lines",
                "PerformanceWarning",
            )

        wavenumber, wavenumber_calc, woutrange = _generate_wavenumber_range(
            self.input.wavenum_min,
            self.input.wavenum_max,
            self.params.wstep,
            neighbour_lines,
        )

        # Generate lineshape array
        if truncation is None:
            # compute lineshape on full range :
            # (note that this means 3x wavenumber_calc will be required when applying lineshapes)
            truncation = wavenumber_calc[-1] - wavenumber_calc[0]

        wbroad_centered = _generate_broadening_range(self.params.wstep, truncation)
        self.truncation = truncation
        # store value for use in lineshape broadening.
        # Note : may be different from self.params.truncation if None was given.

        self.wbroad_centered = wbroad_centered
        self.wavenumber = wavenumber
        self.wavenumber_calc = wavenumber_calc
        self.woutrange = woutrange

        # AccuracyWarning. Check there are enough gridpoints per line.
        self._check_accuracy(self.params.wstep)

        self.profiler.stop("generate_wavenumber_arrays", "Generated Wavenumber Arrays")

        import radis

        if radis.config["DEBUG_MODE"]:
            assert (wavenumber_calc[woutrange[0] : woutrange[1]] == wavenumber).all()

        return

    def predict_time(self):
        def _is_at_equilibrium():
            try:
                assert self.input.Tvib is None or self.input.Tvib == self.input.Tgas
                assert self.input.Trot is None or self.input.Trot == self.input.Tgas
                assert (
                    self.input.overpopulation is None or self.input.overpopulation == {}
                )
                try:
                    if self.input.self_absorption:
                        assert self.input.self_absorption  # == True
                except KeyError:
                    pass
                return True
            except AssertionError:
                return False

        if _is_at_equilibrium() or self.params.optimization is None:
            factor = 1
        else:
            factor = 2  #  _apply_broadening_DLM() is called twice

        wstep = self.params.wstep
        n_lines = self.misc.total_lines
        truncation = self.params.truncation
        spectral_points = (
            self.params.wavenum_max_calc - self.params.wavenum_min_calc
        ) / self.params.wstep

        optimization = self.params.optimization
        broadening_method = self.params.broadening_method

        if optimization in ("simple", "min-RMS"):
            wL = self.wL
            wG = self.wG
            if broadening_method == "voigt":
                estimated_time = (
                    2.096e-07 * n_lines
                    + 7.185e-09
                    * (1 + wL * wG)
                    * spectral_points
                    * np.log(spectral_points)
                    * factor
                )
            elif broadening_method == "fft":
                estimated_time = (
                    4.675e-08
                    * (1 + wL * wG)
                    * spectral_points
                    * np.log(spectral_points)
                    * factor
                )
            elif broadening_method == "convolve":  # Not benchmarked
                estimated_time = (
                    self._broadening_time_ruleofthumb
                    * len(self.df0)
                    * len(self.wbroad_centered)
                )
            else:
                raise NotImplementedError("broadening_method not implemented")
        elif optimization is None:
            if broadening_method == "voigt":
                estimated_time = 6.6487e-08 * n_lines * truncation / wstep * factor
            elif broadening_method == "convolve":  # Not benchmarked
                estimated_time = (
                    self._broadening_time_ruleofthumb
                    * len(self.df0)
                    * len(self.wbroad_centered)
                )
            else:
                raise NotImplementedError("broadening_method not implemented")
        else:
            raise NotImplementedError("optimization not implemented")

        return estimated_time

    def _get_log_2gs(self):
        """Returns log_2gs if it already exists in the dataframe, otherwise
        computes it using gamma_air."""
        df = self.df0
        # TODO: deal with the case of gamma_self [so we don't forget]
        # TODO (refactor) : move into BaseFactory or BroadenFactory (parent classes)

        # if the column already exists, then return
        if "log_2gs" in df.columns:
            return df["log_2gs"]

        try:
            gamma_air = df["airbrd"].to_numpy()
            log_2gs = np.log(2 * gamma_air)
            df["log_2gs"] = log_2gs
            return log_2gs
        except KeyError as err:
            raise KeyError(
                "Cannot find air-broadened half-width or log_2gs in the database... please check the database"
            ) from err

    def _get_log_2vMm(self, molarmass_arr):
        """Returns log_2vMm if it already exists in the dataframe, otherwise
        computes it using the abundance and molar mass for each isotope passed
        in the input."""
        df = self.df0

        # if the column already exists, then return
        if "log_2vMm" in df.columns:
            return df["log_2vMm"]

        try:
            v0 = df["wav"].to_numpy()  # get wavenumber
            iso = df["iso"].to_numpy()  # get isotope
            Mm = molarmass_arr * 1e-3 / N_A
            log_2vMm = np.log(2 * v0) + 0.5 * np.log(
                2 * k * np.log(2) / (c ** 2 * Mm.take(iso))
            )
            df["log_2vMm"] = log_2vMm
            return log_2vMm
        except KeyError as err:
            raise KeyError(
                "Cannot find wavenumber, isotope and/or log_2vMm in the database. Please check the database"
            ) from err

    def _get_S0(self, Ia_arr):
        """Returns S0 if it already exists, otherwise computes the value using
        abundance, gamma_air and Einstein's number."""
        df = self.df0

        # if the column already exists, then return it
        if "S0" in df.columns:
            return df["S0"]

        try:
            v0 = df["wav"].to_numpy()
            iso = df["iso"].to_numpy()
            A21 = df["A"].to_numpy()
            Jl = df["jl"].to_numpy()
            DJ = df["branch"].to_numpy()
            Ju = Jl + DJ
            gu = 2 * Ju + 1  # g_up
            S0 = Ia_arr.take(iso) * gu * A21 / (8 * pi * c_cm * v0 ** 2)
            df["S0"] = S0
            return S0
        except KeyError as err:
            raise KeyError(
                "Could not find wavenumber, Einstein's coefficient, lower state energy or S0 in the dataframe. PLease check the database"
            ) from err

    def optically_thin_power(
        self,
        Tgas=None,
        Tvib=None,
        Trot=None,
        Ttrans=None,
        vib_distribution="boltzmann",
        rot_distribution="boltzmann",
        mole_fraction=None,
        path_length=None,
        unit="mW/cm2/sr",
    ):
        """Calculate total power emitted in equilibrium or non-equilibrium case
        in the optically thin approximation: it sums all emission integral over
        the total spectral range.

        .. warning::

            this is a fast implementation that doesnt take into account
            the contribution of lines outside the given spectral range. It is valid for spectral ranges
            surrounded by no lines, and spectral ranges much broaded than the typical
            line broadening (~ 1-10 cm-1 in the infrared)

        If what you're looking for is an accurate simulation on a narrow spectral range
        you better calculate the spectrum (that does take all of that into account)
        and integrate it with :py:meth:`~radis.spectrum.spectrum.Spectrum.get_power`

        Parameters
        ----------
        Tgas: float
            equilibrium temperature [K]
            If doing a non equilibrium case it should be None. Use Ttrans for
            translational temperature
        Tvib: float
            vibrational temperature [K]
        Trot: float
            rotational temperature [K]
        Ttrans: float
            translational temperature [K]. If None, translational temperature is
            taken as rotational temperature (valid at 1 atm for times above ~ 2ns
            which is the RT characteristic time)
        mole_fraction: float
            database species mole fraction. If None, Factory mole fraction is used.
        path_length: float
            slab size (cm). If None, Factory mole fraction is used.
        unit: str
            output unit. Default ``'mW/cm2/sr'``

        Returns
        -------
        float: Returns total power density in mW/cm2/sr (unless different unit is chosen),
            see ``unit=``.


        See Also
        --------
        :py:meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum`,
        :py:meth:`~radis.spectrum.spectrum.Spectrum.get_power`,
        :py:meth:`~radis.spectrum.spectrum.Spectrum.get_integral`
        """
        # Check inputs

        # ... temperatures

        if Tgas is None and Trot is None:
            raise ValueError(
                "Choose either Tgas (equilibrium) or Tvib / Trot "
                + ". Ttrans (non equilibrium)"
            )

        non_eq_mode = Tgas is None

        if Tvib is None and Trot is not None or Tvib is not None and Trot is None:
            raise ValueError("Choose both Tvib and Trot")

        if non_eq_mode and Ttrans is None:
            Ttrans = Trot

        # update defaults
        if path_length is not None:
            self.input.path_length = path_length
        if mole_fraction is not None:
            self.input.mole_fraction = mole_fraction

        # Get translational temperature
        if Tgas is None:
            Tgas = Ttrans
        self.input.Tgas = Tgas
        self.input.Tvib = Tvib
        self.input.Trot = Trot

        # Init variables
        path_length = self.input.path_length
        mole_fraction = self.input.mole_fraction
        pressure_mbar = self.input.pressure_mbar
        verbose = self.verbose

        # New Profiler object
        self._reset_profiler(verbose)

        self.profiler.start("optically_thin_power_calculation", 1)

        # Make sure database is loaded
        if self.df0 is None:
            if not self.save_memory:
                raise AttributeError("Load databank first (.load_databank())")
            else:
                self._reload_databank()

        if non_eq_mode:
            singleTvibmode = is_float(Tvib)
            # Make sure database has pre-computed non equilibrium quantities
            # (Evib, Erot, etc.)
            self._calc_noneq_parameters(vib_distribution, singleTvibmode)

        # %% Start
        # ----------------------------------------------------------------------

        # Print conditions
        if verbose:
            self.print_conditions(
                "Calculating Radiative Power (optically thin approximation)"
            )

        # Calculate power
        # ---------------------------------------------------

        self._reinitialize()  # creates scaled dataframe df1 from df0

        # ----------------------------------------------------------------------
        # Calculate Populations and Emission Integral
        # (Note: Emission Integral is non canonical quantity, equivalent to
        #  Linestrength for absorption)
        if non_eq_mode:
            self.calc_populations_noneq(
                Tvib,
                Trot,
                vib_distribution=vib_distribution,
                rot_distribution=rot_distribution,
            )
        else:
            self.calc_populations_eq(Tgas)
            self.df1["Aul"] = self.df1.A  # update einstein coefficients
        self.calc_emission_integral()

        #        # ----------------------------------------------------------------------
        #        # Cutoff linestrength  (note that cuting linestrength doesnt make this
        #        # process faster here, but we still give this option to be consistent
        #        # with spectra)
        #        self._cutoff_linestrength()

        # ----------------------------------------------------------------------

        # Sum over all emission integrals (in the valid range)
        b = (self.df1.wav > self.input.wavenum_min) & (
            self.df1.wav < self.input.wavenum_max
        )
        P = self.df1["Ei"][b].sum()  # Ei  >> (mW/sr)

        # incorporate density of molecules (see equation (A.16) )
        density = mole_fraction * ((pressure_mbar * 100) / (k_b * Tgas)) * 1e-6
        Pv = P * density  # (mW/sr/cm3)

        # Optically thin case (no self absorption):
        Ptot = Pv * path_length  # (mW/sr/cm2)

        self.profiler.stop(
            "optically_thin_power_calculation", "Optically thin power calculation"
        )

        return conv2(Ptot, "mW/cm2/sr", unit)

    def fit_spectrum(
        self,
        s_exp,
        model,
        fit_parameters,
        bounds={},
        plot=False,
        solver_options={"maxiter": 300},
        **kwargs,
    ) -> Union[Spectrum, OptimizeResult]:
        """Fit an experimental spectrum with an arbitrary model and an arbitrary
        number of fit parameters.

        Parameters
        ----------
        s_exp : Spectrum
            experimental spectrum. Should have only spectral array only. Use
            :py:meth:`~radis.spectrum.spectrum.Spectrum.take`, e.g::
                sf.fit_spectrum(s_exp.take('transmittance'))
        model : func -> Spectrum
            a line-of-sight model returning a Spectrum. Example :
            :py:func:`~radis.tools.fitting.LTEModel, `:py:func:`~radis.tools.fitting.Tvib12Tvib3Trot_NonLTEModel`
        fit_parameters : dict
            example::

                {fit_parameter:initial_value}
        bounds : dict, optional
            example::

                {fit_parameter:[min, max]}
        fixed_parameters : dict
            fixed parameters given to the model. Example::

                fit_spectrum(fixed_parameters={"vib_distribution":"treanor"})

        Other Parameters
        ----------------
        plot: bool
            if ``True``, plot spectra as they are computed; and plot the convergence of
            the residual. Default ``False``
        solver_options: dict
            parameters forwarded to the solver. More info in `~scipy.optimize.minimize`
            Example::

                {"maxiter": (int)  max number of iteration default ``300``,
                 }
        kwargs: dict
            forwarded to :py:func:`~radis.tools.fitting.fit_spectrum`

        Returns
        -------
        s_best: Spectrum
            best spectrum
        res: OptimizeResults
            output of `~scipy.optimize.minimize`

        Examples
        --------
        See a :ref:`one-temperature fit example <example_one_temperature_fit>`
        and a :ref:`non-LTE fit example <example_multi_temperature_fit>`

        .. minigallery:: radis.lbl.factory.SpectrumFactory.fit_spectrum


        See Also
        --------
        :py:func:`~radis.tools.fitting.fit_spectrum`,
        :py:func:`~radis.tools.fitting.Tvib12Tvib3Trot_NonLTEModel`,
        `For more advanced cases, use Fitroom <https://github.com/radis/fitroom>`

        """
        from radis.tools.fitting import fit_spectrum

        return fit_spectrum(
            self,
            s_exp,
            model,
            fit_parameters,
            bounds=bounds,
            plot=plot,
            solver_options=solver_options,
            **kwargs,
        )

    def print_perf_profile(self, number_format="{:.3f}", precision=16):
        """Prints Profiler output dictionary in a structured manner for
        the last calculated spectrum

        Examples
        --------
        ::

            sf.print_perf_profile()

            # output >>
                spectrum_calculation      0.189s ████████████████
                    check_line_databank              0.000s
                    check_non_eq_param               0.042s ███
                    fetch_energy_5                   0.015s █
                    calc_weight_trans                0.008s
                    reinitialize                     0.002s
                        copy_database                    0.000s
                        memory_usage_warning             0.002s
                        reset_population                 0.000s
                    calc_noneq_population            0.041s ███
                        part_function                    0.035s ██
                        population                       0.006s
                    scaled_non_eq_linestrength       0.005s
                        map_part_func                    0.001s
                        corrected_population_se          0.003s
                    calc_emission_integral           0.006s
                    applied_linestrength_cutoff      0.002s
                    calc_lineshift                   0.001s
                    calc_hwhm                        0.007s
                    generate_wavenumber_arrays       0.001s
                    calc_line_broadening             0.074s ██████
                        precompute_DLM_lineshapes        0.012s
                        DLM_Initialized_vectors          0.000s
                        DLM_closest_matching_line        0.001s
                        DLM_Distribute_lines             0.001s
                        DLM_convolve                     0.060s █████
                        others                           0.001s
                    calc_other_spectral_quan         0.003s
                    generate_spectrum_obj            0.000s
                    others                           -0.016s

        Other Parameters
        ----------------
        precision: int, optional
            total number of blocks. Default 16.

        See Also
        --------

        :py:meth:`~radis.spectrum.spectrum.Spectrum.print_perf_profile`
        """
        from radis.spectrum.utils import print_perf_profile

        profiler = self.profiler.final
        total_time = profiler["spectrum_calculation"]["value"]

        return print_perf_profile(
            profiler, total_time, number_format=number_format, precision=precision
        )

    def generate_perf_profile(self):
        """Generate a visual/interactive performance profile diagram for
        the last calculated spectrum

        .. note:
            requires a `profiler` key with in Spectrum.conditions

        Examples
        --------
        ::

            sf = SpectrumFactory(...)
            sf.eq_spectrum(...)
            sf.generate_perf_profile()

        See typical output in https://github.com/radis/radis/pull/325

        .. image:: https://user-images.githubusercontent.com/16088743/128018032-6049be72-1881-46ac-9d7c-1ed89f9c4f42.png
            :alt: https://user-images.githubusercontent.com/16088743/128018032-6049be72-1881-46ac-9d7c-1ed89f9c4f42.png
            :target: https://user-images.githubusercontent.com/16088743/128018032-6049be72-1881-46ac-9d7c-1ed89f9c4f42.png


        .. note::
            You can also profile with `tuna` directly::

                python -m cProfile -o program.prof your_radis_script.py
                tuna your_radis_script.py


        See Also
        --------
        :py:meth:`~radis.spectrum.spectrum.Spectrum.print_perf_profile`
        """
        from radis.spectrum.utils import generate_perf_profile

        profiler = self.profiler.final.copy().copy()
        # Add total calculation time:
        profiler.update({"value": profiler["spectrum_calculation"]["value"]})
        # note: in Spectrum.generate_perf_profile the total time is taken as
        # 'self.conditions["calculation_time"]' which also includes database
        # loading times.

        return generate_perf_profile(profiler)


# %% ======================================================================
# EXTRA FUNCTIONS
# ---------------------
#
# _generate_wavenumber_range
# _generate_broadening_range
#
# XXX =====================================================================


def _generate_wavenumber_range(wavenum_min, wavenum_max, wstep, neighbour_lines):
    """define waverange vectors, with ``wavenumber`` the output spectral range
    and ``wavenumber_calc`` the spectral range used for calculation, that
    includes neighbour lines within ``neighbour_lines`` distance.

    Parameters
    ----------
    wavenum_min, wavenum_max: float
        wavenumber range limits (cm-1)
    wstep: float
        wavenumber step (cm-1)
    neighbour_lines: float
        wavenumber full width of broadening calculation: used to define which
        neighbour lines shall be included in the calculation

    Returns
    -------
    wavenumber: numpy array
        an evenly spaced array between ``wavenum_min`` and ``wavenum_max`` with
        a spacing of ``wstep``
    wavenumber_calc: numpy array
        an evenly spaced array between ``wavenum_min-neighbour_lines`` and
        ``wavenum_max+neighbour_lines`` with a spacing of ``wstep``
    woutrange: (wmin, wmax)
        index to project the full range including neighbour lines `wavenumber_calc`
        on the final range `wavenumber`, i.e. : wavenumber_calc[woutrange[0]:woutrange[1]] = wavenumber
    """
    assert wavenum_min < wavenum_max

    # Output range
    # generate the final vector of wavenumbers (shape M)
    wavenumber = arange(wavenum_min, wavenum_max + wstep, wstep)

    # generate the calculation vector of wavenumbers (shape M + space on the side)
    # ... Calculation range
    wavenum_min_calc = wavenumber[0] - neighbour_lines  # cm-1
    wavenum_max_calc = wavenumber[-1] + neighbour_lines  # cm-1
    w_out_of_range_left = arange(
        wavenumber[0] - wstep, wavenum_min_calc - wstep, -wstep
    )[::-1]
    w_out_of_range_right = arange(
        wavenumber[-1] + wstep, wavenum_max_calc + wstep, wstep
    )

    # ... deal with rounding errors: 1 side may have 1 more point
    if len(w_out_of_range_left) > len(w_out_of_range_right):
        w_out_of_range_left = w_out_of_range_left[1:]
    elif len(w_out_of_range_left) < len(w_out_of_range_right):
        w_out_of_range_right = w_out_of_range_right[:-1]

    wavenumber_calc = np.hstack((w_out_of_range_left, wavenumber, w_out_of_range_right))
    woutrange = len(w_out_of_range_left), len(w_out_of_range_left) + len(wavenumber)

    assert len(w_out_of_range_left) == len(w_out_of_range_right)
    assert len(wavenumber_calc) == len(wavenumber) + 2 * len(w_out_of_range_left)

    return wavenumber, wavenumber_calc, woutrange


def _generate_broadening_range(wstep, truncation):
    """Generate array on which to compute line broadening.

    Parameters
    ----------
    wstep: float
        wavenumber step (cm-1)
    truncation: float
        wavenumber half-width of broadening calculation: used to define which
        neighbour lines shall be included in the calculation

    Returns
    -------
    wbroad_centered: numpy array
        an evenly spaced array, of odd-parity length, centered on 0, and of width
        ``truncation``
    """

    # create a broadening array, on which lineshape will be calculated.
    # Odd number is important
    wbroad_centered = np.hstack(
        (
            -arange(wstep, truncation + wstep, wstep)[::-1],
            [0],
            arange(wstep, truncation + wstep, wstep),
        )
    )

    assert len(wbroad_centered) % 2 == 1

    return wbroad_centered


# %% Test

# --------------------------
if __name__ == "__main__":

    import pytest

    print("Testing factory:", pytest.main(["../test/lbl/test_factory.py"]))



# -*- coding: utf-8 -*-
"""
@author: Erwan

Logging and warning functions
"""


import sys
from time import gmtime, strftime
from warnings import warn

from termcolor import colored


def printlog(msg, logfile="log.txt", stdout=True):
    """Write a message to the logfile, adding date and time.

    Also print it to screen if stdout is True
    """
    with open(logfile, "a+") as f:
        f.write("{0} {1}\n".format(strftime("%Y-%m-%d %H:%M:%S", gmtime()), msg))
    if stdout:
        print(msg)
    return


def warnlog(msg, logfile="log.txt", stdout=True):
    """Write a WARNING message to the logfile, adding date and time.

    Also print it to screen if stdout is True
    """
    with open(logfile, "a+") as f:
        f.write(
            "{0} WARNING. {1}\n".format(strftime("%Y-%m-%d %H:%M:%S", gmtime()), msg)
        )
    if stdout:
        warn(msg)
    return


def printwarn(msg, verbose=True, warnings=True):
    """Send msg in the console if verbose is True, and as warning if warnings
    is True.

    Warnings is added at the end of the process but you may miss it if
    interrupted.
    """

    if verbose:
        print(colored("\n" + msg + "\n", "red"))
        sys.stdout.flush()
    if warnings:
        warn(msg)
    return



# -*- coding: utf-8 -*-
"""

Examples
--------

Run all tests::

    pytest      # (in command line, in project folder)

Run only fast tests (i.e: tests that a  'fast' label)::

    pytest -m fast

-------------------------------------------------------------------------------


"""

import os
from os.path import basename, exists

import numpy as np
import pytest
from numpy import allclose, linspace

from radis.phys.convert import nm2cm
from radis.spectrum import Spectrum, calculated_spectrum

fig_prefix = basename(__file__) + ": "

# %% Test routines


def test_spectrum_creation_method(*args, **kwargs):
    import pytest

    w = np.linspace(300, 700, 1000)
    k = (np.random.rand(1000) + 0.3) ** 10
    T = np.exp(-k * 1)

    # Good inputs:
    # ... format of quantities :
    Spectrum({"wavelength": w, "abscoeff": k, "transmittance_noslit": T}, wunit="nm")
    Spectrum({"abscoeff": (w, k), "transmittance_noslit": (w, T)}, wunit="nm")

    Spectrum({"wavelength": w, "abscoeff": k, "transmittance_noslit": T}, wunit="nm")

    # ... units:
    Spectrum({"wavespace": w, "abscoeff": k}, wunit="nm")
    Spectrum({"wavelength": w, "abscoeff": k}, wunit="nm")

    # Bad inputs:
    with pytest.raises(AssertionError) as err:
        Spectrum(
            {"wavelength": w, "abscoeff": k, "transmittance_noslit": T}
        )  # wunit not defined
    with pytest.raises(AssertionError) as err:
        Spectrum(
            {"wavenumber": w, "abscoeff": k, "transmittance_noslit": T}
        )  # wunit not defined

    # ... wavespace defined multiple times
    with pytest.raises(AssertionError) as err:
        Spectrum({"wavelength": w, "wavenumber": 1e7 / w, "abscoeff": k}, wunit="nm")
    with pytest.raises(AssertionError) as err:
        Spectrum({"wavelength": w, "wavespace": 1e7 / w, "abscoeff": k}, wunit="nm")

    # ... format of quantities :
    with pytest.raises(AssertionError) as err:
        Spectrum({"wavenumber": w, "abscoeff": np.hstack((k, k))}, wunit="cm-1")
    assert "Input arrays should have the same length" in str(err.value)

    # ... units badly defeined :
    with pytest.raises(AssertionError) as err:
        Spectrum({"wavespace": w, "abscoeff": k})
    assert "waveunit ('nm', 'cm-1'?) has to be defined" in str(err.value)

    with pytest.raises(AssertionError) as err:
        Spectrum({"abscoeff": (w, k)})
    assert "waveunit ('nm', 'cm-1'?) has to be defined" in str(err.value)

    with pytest.raises(AssertionError):
        Spectrum({"wavenumber": w, "abscoeff": k}, wunit="nm")

    with pytest.raises(AssertionError):
        Spectrum({"wavelength": w, "abscoeff": k}, wunit="cm-1")


def test_spectrum_get_methods(
    verbose=True, plot=True, close_plots=True, *args, **kwargs
):
    """Test all spectrum methods on a Spectrum generated in Specair"""

    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec
    from radis.tools.slit import get_FWHM

    if plot and close_plots:
        import matplotlib.pyplot as plt

        plt.close("all")

    s = load_spec(getTestFile("N2C_specair_380nm.spec"), binary=False)

    # general methods
    if verbose:
        print(s)
    dir(s)

    # access properties
    assert s.get_name() == "N2C_specair_380nm"
    assert all(
        s.get_radiance_noslit(Iunit="W/m2/sr/nm")
        == s.get("radiance_noslit", Iunit="W/m2/sr/nm")[1]
    )
    assert all(nm2cm(s.get_wavelength(medium="vacuum")) == s.get_wavenumber())
    assert np.isclose(s.get_power(unit="W/cm2/sr"), 2631.6288408588148)
    assert s.get_waveunit() == "nm"
    assert np.isclose(
        s.get_power(unit="W/cm2/sr"),
        s.get_integral("radiance_noslit", Iunit="W/cm2/sr/nm"),
    )
    assert s.get_conditions()["Tgas"] == 1500
    assert len(s.get_vars()) == 2
    assert s.is_at_equilibrium() == False
    assert s.is_optically_thin() == False

    # Check applied slit has the correct width
    s.apply_slit(0.5)
    wslit, Islit = s.get_slit()
    wstep = np.diff(wslit)[0]
    assert np.isclose(get_FWHM(*s.get_slit()), 0.5, atol=1.1 * wstep)

    if plot:
        s.plot_slit()

    if verbose:
        print("Tested Spectrum methods:")
        print("...print(Spectrum)")
        print(".. get_name()")
        print(".. get_radiance_noslit() vs get()")
        print(".. get_wavelength() vs get_wavenumber")
        print(".. get_power()")
        print(".. get_waveunit()")
        print(".. get_power() vs get_integral()")
        print(".. get_conditions()")
        print(".. get_vars()")
        print(".. is_at_equilibrium()")
        print(".. is_optically_thin()")
        print(".. get_slit()")


@pytest.mark.fast
def test_trimming(verbose=True, *args, **kwargs):
    """Test :py:meth:`radis.spectrum.spectrum.Spectrum.trim`"""

    from radis.misc.arrays import count_nans

    # trim both sides
    w = np.linspace(300, 600)
    T = np.ones_like(w)
    T[:15] = np.nan
    T[-10:] = np.nan
    s = Spectrum.from_array(w, T, "transmittance_noslit", wunit="nm", unit="")

    Nnans = count_nans(T)
    assert len(s) == 50  # before trimming
    s.trim()
    assert len(s) == 50 - Nnans

    # trim one side
    w = np.linspace(300, 600)
    T = np.ones_like(w)
    T[:15] = np.nan
    s = Spectrum.from_array(w, T, "transmittance_noslit", wunit="nm", unit="")

    Nnans = count_nans(T)
    assert len(s) == 50  # before trimming
    s.trim()
    assert len(s) == 50 - Nnans


@pytest.mark.fast
def test_copy(verbose=True, *args, **kwargs):
    """Test that a Spectrum is correctly copied

    We compare a Spectrum that has:
    - all available spectral quantities
    - a slit
    - many calculation conditions
    - no populations
    - no lines
    """

    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"))

    s.update()
    s.apply_slit(1.5)
    s2 = s.copy()

    # Test spectrum in general
    assert s == s2
    assert s is not s2

    # Test all quantities in detail
    for var in s._q.keys():
        assert np.allclose(s._q[var], s2._q[var], equal_nan=True)
        assert not (s._q[var] is s2._q[var])

    if verbose:
        print("Tested that s2 == s (but s2 is not s) after Spectrum copy")


def test_populations(verbose=True, plot=True, close_plots=True, *args, **kwargs):
    """Test that populations in a Spectrum are correctly read"""

    if plot:
        import matplotlib.pyplot as plt

        plt.ion()  # dont get stuck with Matplotlib if executing through pytest

    if plot and close_plots:
        import matplotlib.pyplot as plt

        plt.close("all")

    import pytest

    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec

    # get a spectrum
    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"))

    # Get all populations
    pops = s.get_populations(molecule="CO", isotope=1)
    assert np.isclose(pops["Ia"], 0.986544)

    # Get vib levels
    df_vib = s.get_vib_levels()
    assert len(df_vib) == 49

    # Get rovib levels (dont exist: should fail!)
    with pytest.raises(KeyError):  # expected behavior
        s.get_rovib_levels()

    if plot:
        s.plot_populations()

    if verbose:
        print("test_populations: OK")


def test_store_functions(verbose=True, *args, **kwargs):
    """Test some store / retrieve functions"""

    from radis.spectrum.models import transmittance_spectrum
    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec

    temp_file = "test_radis_tempfile_transmittance.txt"
    assert not exists(temp_file)

    # Test that the transmittance stored on .txt and loaded again match
    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update()
    try:
        s.savetxt(temp_file, "transmittance_noslit", wunit="nm_vac")
        w, T = np.loadtxt(temp_file).T
    finally:
        os.remove(temp_file)

    s2 = transmittance_spectrum(w, T, wunit="nm_vac")
    assert s.compare_with(s2, spectra_only="transmittance_noslit", plot=False)

    # TODO: add test that ensures we can load a binary file without binary=True
    # (and the warning should be captured)

    return True


@pytest.mark.fast
def test_intensity_conversion(verbose=True, *args, **kwargs):
    """Test conversion of intensity cm-1 works:

    - conversion of mW/sr/cm2/nm -> mW/sr/cm2/cm-1

    """

    from radis import planck, planck_wn

    w_nm = linspace(300, 3000)
    w_cm = nm2cm(w_nm)
    I_nm = planck(w_nm, T=6000, unit="mW/sr/cm2/nm")

    s = calculated_spectrum(
        w_nm,
        I_nm,
        wunit="nm_vac",
        Iunit="mW/sr/cm2/nm",
    )
    # mW/sr/cm2/nm -> mW/sr/cm2/cm-1
    w, I = s.get("radiance_noslit", Iunit="mW/sr/cm2/cm-1")
    I_cm = planck_wn(w_cm, T=6000, unit="mW/sr/cm2/cm-1")
    assert allclose(I_cm, I, rtol=1e-3)


# def test_emissivity_conversion(verbose=True, *args, **kwargs):
#    ''' Test conversion of intensity cm-1 works:
#
#    - conversion of mW/sr/cm2/nm -> mW/sr/cm2/cm-1
#
#    '''
#
#    from radis.phys.units import convert_emi2nm
#    from radis.test.utils import getTestFile
#    from radis.tools.database import load_spec
#
#    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'), binary=True)
#    s.update()
#
#    # mW/sr/cm3/nm -> mW/sr/cm3/cm-1
#
#    w, I = s.get('emissivity_noslit', Iunit='mW/sr/cm3/cm-1')
#    I_cm = convert_emi2nm(I, 'mW/sr/cm3/cm-1', 'mW/sr/m3/µm')
#

# TODO: finish implementing emissivity_conversino above


def test_rescaling_function(verbose=True, *args, **kwargs):
    """Test rescaling functions"""

    from radis.test.utils import getTestFile

    s = Spectrum.from_txt(
        getTestFile("calc_N2C_spectrum_Trot1200_Tvib3000.txt"),
        quantity="radiance_noslit",
        wunit="nm",
        unit="mW/cm2/sr/µm",  # Specair units: mW/cm2/sr/µm
        conditions={
            "Tvib": 3000,
            "Trot": 1200,
            "path_length": 1,  # arbitrary
            "medium": "air",
        },
        populations={"molecules": {"N2C": 1e13}},  # arbitrary
        # (just an example)
    )
    s.update(optically_thin=True)
    s.rescale_path_length(10)

    assert np.isclose(s.get_radiance_noslit(Iunit="mW/cm2/sr/nm")[0], 352.57305783248)


def test_resampling_function(
    verbose=True, plot=True, close_plots=True, *args, **kwargs
):
    """Test resampling functions

    Get a Spectrum calculated in cm-1, then resample on a smaller range in cm-1,
    and in approximately the same range (but in nm). Check that all 3 overlap
    """
    # %%
    from radis.spectrum import get_residual_integral
    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec

    if plot and close_plots:
        import matplotlib.pyplot as plt

        plt.close("all")

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.name = "original"

    # Test resampling without changing anything
    s_cm = s.resample(s.get_wavenumber(), "cm-1", inplace=False)
    s_nm = s.resample(s.get_wavelength(), "nm", inplace=False)
    assert s == s_cm
    assert np.allclose(s.get("abscoeff", wunit="nm"), s_nm.get("abscoeff", wunit="nm"))

    # Test resampling on new ranges
    s2 = s.copy()
    s2b = s.copy()
    s3 = s.copy()
    s2.resample(np.linspace(4500, 4700, 10000), unit="nm_vac")
    s2b.resample(np.linspace(4500, 4700, 10000), unit="nm")
    s3.resample(np.linspace(2127.2, 2227.7, 10000), unit="cm-1")
    s2.name = "resampled in nm (vacuum)"
    s2b.name = "resampled in nm (air)"
    s3.name = "resampled in cm-1"

    s.compare_with(
        s2,
        plot=plot,
        title="Residual: {0:.2g}".format(
            get_residual_integral(s, s2, "abscoeff", ignore_nan=True)
        ),
    )
    s.compare_with(
        s2b,
        plot=plot,
        title="Residual: {0:.2g}".format(
            get_residual_integral(s, s2b, "abscoeff", ignore_nan=True)
        ),
    )
    s.compare_with(
        s3,
        plot=plot,
        title="Residual: {0:.2g}".format(
            get_residual_integral(s, s3, "abscoeff", ignore_nan=True)
        ),
    )

    assert get_residual_integral(s, s2, "abscoeff", ignore_nan=True) < 1e-3
    assert get_residual_integral(s, s2b, "abscoeff", ignore_nan=True) < 1e-3
    assert get_residual_integral(s, s3, "abscoeff", ignore_nan=True) < 1e-5


def test_resampling_nan_function(verbose=True, *args, **kwargs):
    """Test resampling functions, for Spectra with nans"""
    from radis import get_residual
    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec

    plot = True

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True).crop(
        2170, 2180, "cm-1"
    )
    s.rescale_path_length(10)  # cm
    s.name = "original"

    # We want to resample on a small range:
    w_exp = s.get_wavenumber()[700:-700][::5]
    # ... add a shift:
    w_exp += np.diff(s.get_wavenumber())[0] * 2 / 3

    s.update("transmittance_noslit")
    s.apply_slit(3, "nm")

    assert s.has_nan()

    # s2.resample(w_exp)
    s2 = s.resample(w_exp, inplace=False)
    s2.name = "resampled"

    assert get_residual(s2, s, "transmittance") < 1e-6

    if plot:
        from radis import plot_diff

        plot_diff(s, s2, show_points=True)


@pytest.mark.fast
def test_noplot_different_quantities(*args, **kwargs):
    """Prevents User Errors: Ensures an error is raised if plotting different
    quantities on the same graph"""

    import matplotlib.pyplot as plt

    plt.ion()

    from radis import load_spec
    from radis.test.utils import getTestFile

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update()
    s.plot("abscoeff", nfig="test_noplot_different_quantities")
    with pytest.raises(ValueError):  # expect an error
        s.plot("emisscoeff", nfig="same")

    plt.close("test_noplot_different_quantities")


@pytest.mark.fast
def test_plot_by_parts(plot=True, *args, **kwargs):
    """Test :py:func:`~radis.misc.plot.split_and_plot_by_parts`
    and plot_by_parts=True in :py:meth:`~radis.spectrum.spectrum.Spectrum.plot`
    """

    import matplotlib.pyplot as plt

    plt.ion()

    from radis import load_spec
    from radis.test.utils import getTestFile

    load_spec(getTestFile("CO2_measured_spectrum_4-5um.spec"), binary=True).plot(
        plot_by_parts=True, nfig="plot by parts (non continuous spectrum)"
    )

    if not plot:
        plt.close("plot by parts (non continuous spectrum)")


@pytest.mark.fast
def test_normalization(*args, **kwargs):

    from radis import Radiance, load_spec
    from radis.test.utils import getTestFile

    # Generate the equivalent of an experimental spectrum
    s = load_spec(getTestFile(r"CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update()  # add radiance, etc.
    s.apply_slit(0.5)  # nm
    s = Radiance(s)

    # Test normalization
    assert s.units["radiance"] != ""
    s.normalize()
    assert s.max() != 1
    s.normalize(inplace=True)
    assert s.max() == 1
    assert s.normalize().units["radiance"] == ""

    s2 = s.normalize(normalize_how="area")
    assert np.isclose(s2.get_integral("radiance", wunit=s2.get_waveunit()), 1)

    #
    s3 = s.normalize(wrange=((2125, 2150)), normalize_how="area")
    assert np.isclose(
        s3.crop(2125, 2150).get_integral("radiance", wunit=s3.get_waveunit()), 1
    )


@pytest.mark.fast
def test_sort(*args, **kwargs):
    """Test :py:meth:`~radis.spectrum.spectrum.Spectrum.sort`"""

    from radis import load_spec
    from radis.test.utils import getTestFile

    s_exp = load_spec(getTestFile("CO2_measured_spectrum_4-5um.spec"), binary=True)

    from radis.misc.arrays import is_sorted

    assert not is_sorted(s_exp.get("radiance")[0])
    assert is_sorted(s_exp.sort().get("radiance")[0])


# %%


def _run_testcases(
    plot=True,
    close_plots=False,
    verbose=True,
    debug=False,
    warnings=True,
    *args,
    **kwargs
):
    """Test procedures

    Parameters
    ----------

    debug: boolean
        swamps the console namespace with local variables. Default ``False``

    """

    # Test all Spectrum methods
    # -------------------------
    test_spectrum_creation_method(*args, **kwargs)
    test_spectrum_get_methods(
        debug=debug,
        verbose=verbose,
        plot=plot,
        close_plots=close_plots,
        *args,
        **kwargs
    )
    test_copy(verbose=verbose, *args, **kwargs)
    test_trimming(*args, **kwargs)
    test_populations(
        verbose=verbose, plot=plot, close_plots=close_plots, *args, **kwargs
    )
    test_store_functions(verbose=verbose, *args, **kwargs)

    # Test populations
    # ----------
    test_populations(verbose=verbose, *args, **kwargs)

    # Test conversion of intensity cm-1 works
    # -------------
    test_intensity_conversion(debug=debug, verbose=verbose, *args, **kwargs)

    # Test updating / rescaling functions (no self absorption)
    # ---------
    test_rescaling_function(debug=debug, *args, **kwargs)

    test_resampling_function(
        debug=debug, plot=plot, close_plots=close_plots, *args, **kwargs
    )
    test_resampling_nan_function(verbose=verbose, *args, **kwargs)

    test_normalization(*args, **kwargs)

    # Test plot firewalls:
    test_noplot_different_quantities(*args, **kwargs)

    # Test plot by parts
    test_plot_by_parts(plot=plot, *args, **kwargs)
    test_sort()

    return True


if __name__ == "__main__":
    print(("Test spectrum: ", _run_testcases(debug=False, close_plots=False)))



# -*- coding: utf-8 -*-
"""
Summary
-------

Public (front-end) functions to calculate Spectrum with HITRAN / CDSD databanks.
Uses the SpectrumFactory class from `factory.py`, Spectrum from `spectrum.py`
and line survey from `line_survey.py`

Routine Listing
---------------

:func:`~radis.lbl.calc.calc_spectrum`

-------------------------------------------------------------------------------

"""


from copy import deepcopy
from os.path import exists

try:  # Proper import
    from .labels import SpectrumFactory
except ImportError:  # if ran from here
    from radis.lbl.factory import SpectrumFactory
from radis.misc.basics import all_in
from radis.misc.utils import Default
from radis.phys.air import air2vacuum
from radis.phys.convert import nm2cm
from radis.spectrum.spectrum import Spectrum


# %%
def calc_spectrum(
    wavenum_min=None,
    wavenum_max=None,
    wavelength_min=None,
    wavelength_max=None,
    Tgas=None,
    Tvib=None,
    Trot=None,
    pressure=1.01325,
    molecule=None,
    isotope="all",
    mole_fraction=1,
    path_length=1,
    databank="hitran",
    medium="air",
    wstep=0.01,
    truncation=Default(50),
    neighbour_lines=0,
    cutoff=1e-27,
    parsum_mode="full summation",
    optimization="min-RMS",
    broadening_method=Default("voigt"),
    overpopulation=None,
    name=None,
    save_to="",
    use_cached=True,
    mode="cpu",
    export_lines=False,
    verbose=True,
    **kwargs
) -> Spectrum:
    r"""Multipurpose function to calculate a :py:class:`~radis.spectrum.spectrum.Spectrum`.

    Can automatically download databases (HITRAN/HITEMP) or use manually downloaded
    local databases, under equilibrium or non-equilibrium, with or without overpopulation,
    using either CPU or GPU.

    It is a wrapper to :py:class:`~radis.lbl.factory.SpectrumFactory` class.
    For advanced used, please refer to the aforementionned class.


    Parameters
    ----------
    wavenum_min, wavenum_max: float [:math:`cm^{-1}`] or `~astropy.units.quantity.Quantity`
        wavenumber range to be processed in :math:`cm^{-1}`. Use arbitrary units::

            import astropy.units as u
            calc_spectrum(2.5*u.um, 3.0*u.um, ...)

    wavelength_min, wavelength_max: float [:math:`nm`] or `~astropy.units.quantity.Quantity`
        wavelength range to be processed in :math:`nm`. Wavelength in ``'air'`` or
        ``'vacuum'`` depending of the value of ``'medium'``.
    Tgas: float [:math:`K`]
        Gas temperature. If non equilibrium, is used for :math:`T_{translational}`.
        Default ``300`` K​
    Tvib, Trot: float [:math:`K`]
        Vibrational and rotational temperatures (for non-LTE calculations).
        If ``None``, they are at equilibrium with ``Tgas`` ​.
    pressure: float [:math:`bar`] or `~astropy.units.quantity.Quantity`
        partial pressure of gas in bar. Default ``1.01325`` (1 atm)​. Use arbitrary units::

            import astropy.units as u
            calc_spectrum(..., pressure=20*u.mbar)

    molecule: int, str, list or ``None``
        molecule id (HITRAN format) or name. For multiple molecules, use a list.
        The ``'isotope'``, ``'mole_fraction'``, ``'databank'`` and ``'overpopulation'`` parameters must then
        be dictionaries.
        If ``None``, the molecule can be infered
        from the database files being loaded. See the list of supported molecules
        in :py:data:`~radis.db.MOLECULES_LIST_EQUILIBRIUM`
        and :py:data:`~radis.db.MOLECULES_LIST_NONEQUILIBRIUM`.
        Default ``None``.​
    isotope: int, list, str of the form ``'1,2'``, or ``'all'``, or dict
        isotope id (sorted by relative density: (eg: 1: CO2-626, 2: CO2-636 for CO2).
        See [HITRAN-2016]_ documentation for isotope list for all species. If ``'all'``,
        all isotopes in database are used (this may result in larger computation
        times!). Default ``'all'``.

        For multiple molecules, use a dictionary with molecule names as keys ::

            isotope={'CO2':'1,2' ,  'CO':'1,2,3' }​

    mole_fraction: float or dict
        database species mole fraction. Default ``1``.

        For multiple molecules, use a dictionary with molecule names as keys ::

            mole_fraction={'CO2': 0.8, 'CO':0.2}​

    path_length: float [:math:`cm`] or `~astropy.units.quantity.Quantity`
        slab size. Default ``1`` cm​. Use arbitrary units::

            import astropy.units as u
            calc_spectrum(..., path_length=1000*u.km)

    databank: str or dict
        can be either:
        - ``'hitran'``, to fetch automatically the latest HITRAN version
          through :py:func:`~radis.io.query.fetch_astroquery` (based on [HAPI]_ ).
        - ``'hitemp'``, to fetch automatically the latest HITEMP version
          through :py:func:`~radis.io.hitemp.fetch_hitemp`.
        - the name of a a valid database file, in which case the format is inferred.
          For instance, ``'.par'`` is recognized as ``hitran/hitemp`` format.
          Accepts wildcards ``'*'`` to select multiple files.
        - the name of a spectral database registered in your ``~/radis.json``
          :ref:`configuration file <label_lbl_config_file>`.
          This allows to use multiple database files.

        Default ``'hitran'``. See :class:`~radis.lbl.loader.DatabankLoader` for more
        information on line databases, and :data:`~radis.misc.config.DBFORMAT` for
        your ``~/radis.json`` file format.

        For multiple molecules, use a dictionary with molecule names as keys::

            databank='hitran'     # automatic download (or 'hitemp')
            databank='PATH/TO/05_HITEMP2019.par'    # path to a file
            databank='*CO2*.par' #to get all the files that have CO2 in their names (case insensitive)
            databank='HITEMP-2019-CO'   # user-defined database in Configuration file
            databank = {'CO2' : 'PATH/TO/05_HITEMP2019.par', 'CO' : 'hitran'}  # for multiple molecules

    Other Parameters
    ----------------
    medium: ``'air'``, ``'vacuum'``
        propagating medium when giving inputs with ``'wavenum_min'``, ``'wavenum_max'``.
        Does not change anything when giving inputs in wavenumber. Default ``'air'``​ .
    wstep: float (:math:`cm^{-1}`)  or `'auto'`
        Resolution of wavenumber grid. Default ``0.01`` cm-1.
        If `'auto'`, it is ensured that there
        are slightly more or less than :py:data:`~radis.params.GRIDPOINTS_PER_LINEWIDTH_WARN_THRESHOLD`
        points for each linewidth.

        .. note::
            wstep = 'auto' is optimized for performances while ensuring accuracy,
            but is still experimental in 0.9.30. Feedback welcome!
    truncation: float (:math:`cm^{-1}`)
        Half-width over which to compute the lineshape, i.e. lines are truncated
        on each side after ``truncation`` (:math:`cm^{-1}`) from the line center.
        If ``None``, use no truncation (lineshapes spread on the full spectral range).
        Default is ``300`` :math:`cm^{-1}`

        .. note::
                Large values (> ``50``) can induce a performance drop (computation of lineshape
                typically scale as :math:`~truncation ^2` ). The default ``300`` was
                chosen to maintain a good accuracy, and still exhibit the sub-Lorentzian
                behavior of most lines far (few hundreds :math:`cm^{-1}`) from the line center.
    neighbour_lines: float (:math:`cm^{-1}`)
        The calculated spectral range is increased (by ``neighbour_lines`` cm-1
        on each side) to take into account overlaps from out-of-range lines.
        Default is ``0`` :math:`cm^{-1}`.​
    cutoff: float (~ unit of Linestrength: :math:`cm^{-1}/(molec.cm^{-2})`)
        discard linestrengths that are lower that this, to reduce calculation
        times. ``1e-27`` is what is generally used to generate line databases such as
        CDSD. If ``0``, no cutoff. Default ``1e-27`` .
    parsum_mode: 'full summation', 'tabulation'
        how to compute partition functions, at nonequilibrium or when partition
        function are not already tabulated. ``'full summation'`` : sums over all
        (potentially millions) of rovibrational levels. ``'tabulation'`` :
        builds an on-the-fly tabulation of rovibrational levels (500 - 4000x faster
        and usually accurate within 0.1%). Default ``full summation'``

        .. note::
            parsum_mode= 'tabulation'  is new in 0.9.30, and makes nonequilibrium
            calculations of small spectra extremelly fast. Will become the default
            after 0.9.31.
    optimization : ``"simple"``, ``"min-RMS"``, ``None``
        If either ``"simple"`` or ``"min-RMS"`` LDM optimization for lineshape calculation is used:

        - ``"min-RMS"`` : weights optimized by analytical minimization of the RMS-error (See: [Spectral-Synthesis-Algorithm]_)
        - ``"simple"`` : weights equal to their relative position in the grid

        If using the LDM optimization, broadening method is automatically set to ``'fft'``.
        If ``None``, no lineshape interpolation is performed and the lineshape of all lines is calculated.
        Refer to [Spectral-Synthesis-Algorithm]_ for more explanation on the LDM method for lineshape interpolation.
        Default ``"min-RMS"``.
    overpopulation: dict
        dictionary of overpopulation compared to the given vibrational temperature.
        Default ``None``. Example::

            overpopulation = {'CO2' : {'(00`0`0)->(00`0`1)': 2.5,
                                       '(00`0`1)->(00`0`2)': 1,
                                       '(01`1`0)->(01`1`1)': 1,
                                       '(01`1`1)->(01`1`2)': 1 }
                             }​
    export_lines: boolean
        if ``True``, saves details of all calculated lines in Spectrum. This is
        necessary to later use :py:meth:`~radis.spectrum.spectrum.Spectrum.line_survey`,
        but can take some space. Default ``False``.
    name: str
        name of the output Spectrum. If ``None``, a unique ID is generated.
    save_to: str
        save to a `.spec` file which contains absorption & emission features, all
        calculation parameters, and can be opened with :py:func:`~radis.tools.database.load_spec`.
        File can be reloaded and exported to text formats afterwards, see
        :py:meth:`~radis.spectrum.spectrum.Spectrum.savetxt`.
        If file already exists, replace.
    use_cached: boolean
        use cached files for line database and energy database. Default ``True``.​
    verbose: boolean, or int
        If ``False``, stays quiet. If ``True``, tells what is going on.
        If ``>=2``, gives more detailed messages (for instance, details of
        calculation times). Default ``True``.​
    mode: ``'cpu'``, ``'gpu'``
        if set to ``'cpu'``, computes the spectra purely on the CPU. if set to ``'gpu'``,
        offloads the calculations of lineshape and broadening steps to the GPU
        making use of parallel computations to speed up the process. Default ``'cpu'``.
        Note that ``mode='gpu'`` requires CUDA compatible hardware to execute.
        For more information on how to setup your system to run GPU-accelerated
        methods using CUDA and Cython, check `GPU Spectrum Calculation on RADIS <https://radis.readthedocs.io/en/latest/lbl/gpu.html>`__
    **kwargs: other inputs forwarded to SpectrumFactory
        For instance: ``warnings``.
        See :class:`~radis.lbl.factory.SpectrumFactory` documentation for more
        details on input.

    Returns
    -------
    :class:`~radis.spectrum.spectrum.Spectrum`
        Output spectrum:

        - Use the :py:meth:`~radis.spectrum.spectrum.Spectrum.get` method to retrieve a
          spectral quantity (``'radiance'``, ``'radiance_noslit'``, ``'absorbance'``, etc...)
        - Or the :py:meth:`~radis.spectrum.spectrum.Spectrum.plot` method to plot it
          directly.
        - See [1]_ to get an overview of all Spectrum methods

    References
    ----------
    .. [1] RADIS doc: `Spectrum how to? <https://radis.readthedocs.io/en/latest/spectrum/spectrum.html#label-spectrum>`__

    ​.. [2] RADIS GPU support: `GPU Calculations on RADIS <https://radis.readthedocs.io/en/latest/lbl/gpu.html>`__
    ​
    Examples
    --------
    Calculate a CO spectrum from the HITRAN database::

        from radis import calc_spectrum
        s = calc_spectrum(1900, 2300,         # cm-1
                          molecule='CO',
                          isotope='1,2,3',
                          pressure=1.01325,   # bar
                          Tgas=1000,
                          mole_fraction=0.1,
                          databank='hitran',  # or 'hitemp'
                          )
        s.apply_slit(0.5, 'nm')
        s.plot('radiance')

    This example uses the :py:meth:`~radis.spectrum.spectrum.Spectrum.apply_slit`
    and :py:meth:`~radis.spectrum.spectrum.Spectrum.plot` methods. See also
    :py:meth:`~radis.spectrum.spectrum.Spectrum.line_survey`::

        s.line_survey(overlay='radiance')

    Calculate a CO2 spectrum from the CDSD-4000 database::

        s = calc_spectrum(2200, 2400,   # cm-1
                          molecule='CO2',
                          isotope='1',
                          databank='/path/to/cdsd/databank/in/npy/format/',
                          pressure=0.1,  # bar
                          Tgas=1000,
                          mole_fraction=0.1,
                          mode='gpu'
                          )

        s.plot('absorbance')

    This example uses the :py:meth:`~radis.lbl.factory.SpectrumFactory.eq_spectrum_gpu` method to calculate
    the spectrum on the GPU. The databank points to the CDSD-4000 databank that has been
    pre-processed and stored in ``numpy.npy`` format.
    ​
    Refer to the online :ref:`Examples <label_examples>` for more cases, and to
    the :ref:`Spectrum page <label_spectrum>` for details on post-processing methods.

    For more details on how to use the GPU method and process the database, refer to the examples
    linked above and the documentation on :ref:`GPU support for RADIS <label_radis_gpu>`.
    ​
    .. minigallery:: radis.lbl.calc.calc_spectrum
        :add-heading:

    References
    ----------
    **cite**: RADIS is built on the shoulders of many state-of-the-art packages and databases. If using RADIS
    to compute spectra, make sure you cite all of them, for proper reproducibility and acknowledgement of
    the work ! See :ref:`How to cite? <label_cite>`

    See Also
    --------
    :py:class:`~radis.lbl.factory.SpectrumFactory`
    """
    from radis.los.slabs import MergeSlabs

    if molecule is not None and type(molecule) != list:
        molecule = [molecule]  # fall back to the other case: multiple molecules

    # Stage 1. Find all molecules, whatever the user input configuration

    # ... Input arguments that CAN be dictionaries of molecules.
    DICT_INPUT_ARGUMENTS = {
        "isotope": isotope,
        "mole_fraction": mole_fraction,
        "databank": databank,
    }
    # Same, but when the values of the arguments themselves are already a dict.
    # (dealt with separately because we cannot use them to guess what are the input molecules)
    DICT_INPUT_DICT_ARGUMENTS = {"overpopulation": overpopulation}

    def _check_molecules_are_consistent(
        molecule_reference_set, reference_name, new_argument, new_argument_name
    ):
        """Will test that molecules set are the same in molecule_reference_set
        and new_argument, if new_argument is a dict. molecule_reference_set is
        a set of molecules (yeah!). reference_name is the name of the argument
        from which we guessed the list of molecules (used to have a clear error
        message). new_argument is the new argument to check new_argument_name
        is its name.

        Returns the set of molecules as found in new_argument, if applicable, else the molecule_reference_set (this allows us to parse all arguments too)

        Note that names are just here to provide clear error messages to the user if there is a contradiction.
        """
        if isinstance(new_argument, dict):
            if molecule_reference_set is None:  # input molecules are still unknown
                return set(new_argument.keys()), new_argument_name
            elif set(new_argument.keys()) != set(molecule_reference_set):
                raise ValueError(
                    "Keys of molecules in the {0} dictionary must be the same as given in `{1}=`, i.e: {2}. Instead, we got {3}".format(
                        new_argument_name,
                        reference_name,
                        molecule_reference_set,
                        set(new_argument.keys()),
                    )
                )
            else:
                return (
                    set(new_argument.keys()),
                    new_argument_name,
                )  # so now we changed the reference
        else:
            return molecule_reference_set, reference_name

    # Parse all inputs:
    molecule_reference_set = molecule
    reference_name = "molecule"
    for argument_name, argument in DICT_INPUT_ARGUMENTS.items():
        molecule_reference_set, reference_name = _check_molecules_are_consistent(
            molecule_reference_set, reference_name, argument, argument_name
        )

    # ... Now we are sure there are no contradctions. Just ensure we have molecules:
    if molecule_reference_set is None:
        raise ValueError(
            "Please enter the molecule(s) to calculate in the `molecule=` argument or as a dictionary in the following: {0}".format(
                list(DICT_INPUT_ARGUMENTS.keys())
            )
        )

    # Stage 2. Now we have the list of molecules. Let's get the input arguments for each of them.

    # ... Initialize and fill the master-dictionary
    molecule_dict = {}

    for molecule in molecule_reference_set:
        molecule_dict[molecule] = {}

        for argument_name, argument_dict in DICT_INPUT_ARGUMENTS.items():
            if isinstance(argument_dict, dict):
                # Choose the correspond value
                molecule_dict[molecule][argument_name] = argument_dict[molecule]
                # Will raise a KeyError if not defined. That's fine!
                # TODO: maybe need to catch KeyError and raise a better error message?
            else:  # argument_name is not a dictionary.
                # Let's distribute the same value to every molecule:
                molecule_dict[molecule][argument_name] = argument_dict
                # If wrong argument, it will be caught in _calc_spectrum() later.

    # ... Special case of dictionary arguments. Find out if they were given as default, or per dictionary of molecules:
    is_same_for_all_molecules = dict.fromkeys(DICT_INPUT_DICT_ARGUMENTS)
    for argument_name, argument_dict in DICT_INPUT_DICT_ARGUMENTS.items():
        if not isinstance(argument_dict, dict):
            is_same_for_all_molecules[argument_name] = True
        else:
            # Argument is a dictionary. Guess if keys are molecules, or levels.
            # Ex: overpopulation dict could be {'CO2':{'(0,0,0,1)':10}} or directly {{'(0,0,0,1)':10}}
            argument_keys = set(argument_dict.keys())
            if all_in(argument_keys, molecule_reference_set):
                is_same_for_all_molecules[argument_name] = False
            else:
                is_same_for_all_molecules[argument_name] = True
    # ... now fill them in:
    for argument_name, argument_dict in DICT_INPUT_DICT_ARGUMENTS.items():
        if is_same_for_all_molecules[argument_name]:
            for mol in molecule_reference_set:
                # copy the value for everyone
                molecule_dict[mol][argument_name] = deepcopy(
                    argument_dict
                )  # in case it gets edited.
        else:  # argument_dict keys are the molecules:
            for mol in molecule_reference_set:
                molecule_dict[mol][argument_name] = argument_dict[mol]

    # Stage 3: Now let's calculate all the spectra
    s_list = []
    for molecule, dict_arguments in molecule_dict.items():
        kwargs_molecule = deepcopy(
            kwargs
        )  # these are the default supplementary arguments. Deepcopy ensures that they remain the same for all molecules, even if modified in _calc_spectrum

        # We add all of the DICT_INPUT_ARGUMENTS values:
        kwargs_molecule.update(**dict_arguments)

        s_list.append(
            _calc_spectrum(
                wavenum_min=wavenum_min,
                wavenum_max=wavenum_max,
                wavelength_min=wavelength_min,
                wavelength_max=wavelength_max,
                Tgas=Tgas,
                Tvib=Tvib,
                Trot=Trot,
                pressure=pressure,
                # overpopulation=overpopulation,  # now in dict_arguments
                molecule=molecule,
                # isotope=isotope,                # now in dict_arguments
                # mole_fraction=mole_fraction,    # now in dict_arguments
                path_length=path_length,
                # databank=databank,              # now in dict_arguments
                medium=medium,
                wstep=wstep,
                truncation=truncation,
                neighbour_lines=neighbour_lines,
                cutoff=cutoff,
                parsum_mode=parsum_mode,
                optimization=optimization,
                broadening_method=broadening_method,
                name=name,
                use_cached=use_cached,
                verbose=verbose,
                mode=mode,
                export_lines=export_lines,
                **kwargs_molecule
            )
        )

    # Stage 4: merge all molecules and return
    s = MergeSlabs(*s_list)

    if save_to:
        s.store(path=save_to, if_exists_then="replace", verbose=verbose)

    return s


def _calc_spectrum(
    wavenum_min,
    wavenum_max,
    wavelength_min,
    wavelength_max,
    Tgas,
    Tvib,
    Trot,
    pressure,
    overpopulation,
    molecule,
    isotope,
    mole_fraction,
    path_length,
    databank,
    medium,
    wstep,
    truncation,
    neighbour_lines,
    cutoff,
    parsum_mode,
    optimization,
    broadening_method,
    name,
    use_cached,
    verbose,
    mode,
    export_lines,
    **kwargs
):
    """See :py:func:`~radis.lbl.calc.calc_spectrum`"""

    # Check inputs

    # ... wavelengths / wavenumbers
    if (wavelength_min is not None or wavelength_max is not None) and (
        wavenum_min is not None or wavenum_max is not None
    ):
        raise ValueError("Wavenumber and Wavelength both given... it's time to choose!")

    if not medium in ["air", "vacuum"]:
        raise NotImplementedError("Unknown propagating medium: {0}".format(medium))

    if wavenum_min is None and wavenum_max is None:
        assert wavelength_max is not None
        assert wavelength_min is not None
        if medium == "vacuum":
            wavenum_min = nm2cm(wavelength_max)
            wavenum_max = nm2cm(wavelength_min)
        else:
            wavenum_min = nm2cm(air2vacuum(wavelength_max))
            wavenum_max = nm2cm(air2vacuum(wavelength_min))
    else:
        assert wavenum_min is not None
        assert wavenum_max is not None

    # ... temperatures

    if Tgas is None and Trot is None:
        raise ValueError(
            "Choose either Tgas (equilibrium) or Tvib / Trot (non equilibrium)"
        )

    if Tvib is None and Trot is not None or Tvib is not None and Trot is None:
        raise ValueError("Choose both Tvib and Trot")

    # ... others
    if databank is None:
        raise ValueError("Give a databank name")

    if not "save_memory" in kwargs:
        # no need to save intermediary results as
        # factory is used once only
        kwargs["save_memory"] = True

    if "chunksize" in kwargs:
        raise DeprecationWarning("use optimization= instead of chunksize=")

    def _is_at_equilibrium():
        try:
            assert Tvib is None or Tvib == Tgas
            assert Trot is None or Trot == Tgas
            assert overpopulation is None
            if "self_absorption" in kwargs:
                assert kwargs["self_absorption"]  # == True
            return True
        except AssertionError:
            return False

    _equilibrium = _is_at_equilibrium()

    # which columns to keep when loading line database
    if kwargs["save_memory"] >= 2 and _equilibrium:
        drop_columns = "all"
    else:
        drop_columns = "auto"

    # Run calculations
    sf = SpectrumFactory(
        wavenum_min,
        wavenum_max,
        medium=medium,
        molecule=molecule,
        isotope=isotope,
        pressure=pressure,
        wstep=wstep,
        truncation=truncation,
        neighbour_lines=neighbour_lines,
        cutoff=cutoff,
        parsum_mode=parsum_mode,
        verbose=verbose,
        optimization=optimization,
        broadening_method=broadening_method,
        export_lines=export_lines,
        **kwargs
    )
    if databank in ["fetch", "hitran", "hitemp", "exomol",] or (
        isinstance(databank, tuple) and databank[0] == "exomol"
    ):  # mode to get databank without relying on  Line databases
        # Line database :
        if databank in ["fetch", "hitran"]:
            conditions = {
                "source": "hitran",
                "parfuncfmt": "hapi",  # use HAPI (TIPS) partition functions for equilibrium
            }
        elif databank in ["hitemp"]:
            conditions = {
                "source": "hitemp",
                "parfuncfmt": "hapi",  # use HAPI (TIPS) partition functions for equilibrium}
            }
        elif databank in ["exomol"]:
            conditions = {
                "source": "exomol",
                "parfuncfmt": "exomol",  # download & use Exo partition functions for equilibrium}
            }
        elif isinstance(databank, tuple) and databank[0] == "exomol":
            conditions = {
                "source": "exomol",
                "exomol_database": databank[1],
                "parfuncfmt": "exomol",  # download & use Exo partition functions for equilibrium}
            }
        # Partition functions :
        conditions.update(
            **{
                "levelsfmt": None,  # no need to load energies by default
                "db_use_cached": use_cached,
            }
        )
        # Rovibrational energies :
        if not _equilibrium:
            # calculate partition functions with energy levels from built-in
            # constants (not all molecules are supported!)
            conditions["levelsfmt"] = "radis"
            conditions["lvl_use_cached"] = use_cached
        conditions["load_energies"] = not _equilibrium
        # Details to identify lines
        conditions["parse_local_global_quanta"] = (not _equilibrium) or export_lines
        sf.fetch_databank(**conditions)
    elif exists(databank):
        conditions = {
            "path": databank,
            "drop_columns": drop_columns,
            "parfuncfmt": "hapi",  # use HAPI (TIPS) partition functions for equilibrium
            "levelsfmt": None,  # no need to load energies by default
            "db_use_cached": use_cached,
        }
        # Guess format
        if databank.endswith(".par"):
            if verbose:
                print("Infered {0} is a HITRAN-format file.".format(databank))
            conditions["format"] = "hitran"
            # If non-equilibrium we'll also need to load the energy levels.
            if not _equilibrium:
                # calculate partition functions with energy levels from built-in
                # constants (not all molecules are supported!)
                conditions["levelsfmt"] = "radis"
                conditions["lvl_use_cached"] = use_cached
        elif databank.endswith(".h5") or databank.endswith(".hdf5"):
            if verbose:
                print(
                    "Infered {0} is a HDF5 file with RADISDB columns format".format(
                        databank
                    )
                )
            conditions["format"] = "hdf5-radisdb"
            if not _equilibrium:
                conditions["levelsfmt"] = "radis"
        else:
            raise ValueError(
                "Couldnt infer the format of the line database file: {0}. ".format(
                    databank
                )
                + "Create a user-defined database in your ~/radis.json file "
                + "and define the format there. More information on "
                + "https://radis.readthedocs.io/en/latest/lbl/lbl.html#configuration-file"
            )
        conditions["load_energies"] = not _equilibrium
        sf.load_databank(**conditions)

    else:  # manual mode: get from user-defined line databases defined in ~/radis.json
        sf.load_databank(
            databank,
            load_energies=not _equilibrium,  # no need to load/calculate energies at eq.
            drop_columns=drop_columns,
        )

    #    # Get optimisation strategies
    #    if lineshape_optimization == 'auto':        # NotImplemented: finally we use DLM all the time as default.
    #        if len(sf.df0) > 1e5:
    #            lineshape_optimization = 'DLM'
    #        else:
    #            lineshape_optimization = None
    #        sf.params['chunksize'] = lineshape_optimization

    if overpopulation is not None or overpopulation != {}:
        sf.misc.export_rovib_fraction = True  # required to compute Partition fucntions with overpopulation being taken into account

    # Use the standard eq_spectrum / non_eq_spectrum functions
    if _equilibrium:
        if mode == "cpu":
            s = sf.eq_spectrum(
                Tgas=Tgas,
                mole_fraction=mole_fraction,
                path_length=path_length,
                name=name,
            )
        elif mode == "gpu":
            s = sf.eq_spectrum_gpu(
                Tgas=Tgas,
                mole_fraction=mole_fraction,
                pressure=pressure,
                path_length=path_length,
                name=name,
            )
        else:
            raise ValueError(mode)
    else:
        if mode != "cpu":
            raise NotImplementedError(mode)
        s = sf.non_eq_spectrum(
            Tvib=Tvib,
            Trot=Trot,
            Ttrans=Tgas,
            overpopulation=overpopulation,
            mole_fraction=mole_fraction,
            path_length=path_length,
            name=name,
        )

    return s


# --------------------------
if __name__ == "__main__":

    from radis.test.lbl.test_calc import _run_testcases

    print(_run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""
Created on Mon May  7 17:34:52 2018

@author: erwan
"""

from os.path import exists
from shutil import rmtree

import matplotlib.pyplot as plt
import pytest

from radis.lbl import SpectrumFactory
from radis.misc.printer import printm
from radis.test.utils import getTestFile, setup_test_line_databases


def test_retrieve_from_database(
    plot=False, verbose=True, warnings=True, *args, **kwargs
):
    """Test autoretrieve from a database:

    first generate an empty :py:class:`~radis.tools.database.SpecDatabase`
    associated to a :py:class`~radis.lbl.factory.SpectrumFactory`,
    then calculate a first spectrum, then calculate it again and make sure
    it is retrieved from the database
    """

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    temp_database_name = "temp_spec_database"

    try:
        if verbose:
            printm(">>> test_retrieve_from_database")

        assert not exists(temp_database_name)

        setup_test_line_databases()  # add HITEMP-CO2-TEST in ~/radis.json if not there

        sf = SpectrumFactory(
            2284.2,
            2284.6,
            wstep=0.001,  # cm-1
            pressure=20 * 1e-3,  # bar
            cutoff=0,
            path_length=0.1,
            mole_fraction=400e-6,
            molecule="CO2",
            isotope=[1],
            medium="vacuum",
            truncation=5,
            export_populations="rovib",
            verbose=verbose,
        )
        sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
        sf.init_databank(
            "HITEMP-CO2-TEST"
        )  # unlike load_databank, will automatically be built when needed
        db = sf.init_database(temp_database_name, autoretrieve=True)

        assert len(db) == 0

        # Calculate a first spectrum
        s1 = sf.non_eq_spectrum(2000, 1000)

        assert len(db) == 1
        # Note that the Spectrum in database is not s1 because populations
        # are not stored by default

        # hack: now change the `autoretrieve` status to `force` to make sure
        # the spectrum is retrieved. Else, an error will be raised
        sf.autoretrievedatabase = "force"

        # Calculate spectrum under the same conditions
        s2 = sf.non_eq_spectrum(2000, 1000)

        # Note that s1 == s2 won't work because populations are not stored
        # by default in the database
        assert s1.compare_with(s2, spectra_only=True, plot=plot)

        return True

    finally:
        rmtree(temp_database_name)


@pytest.mark.fast
def test_ignore_cached_files():
    """
    Previous implementation of RADIS saved the cached h5 files generated while reading the
    dataset in the same directory from where the data was being read. Using a wildcard input
    such as `path = "cdsd_hitemp_09_frag*"` in such case led to the cached files present
    in directory to also being loaded and treated as the dataset files. This resulted in
    an error due to the differences in the way data is stored in h5 files versus in dataset
    files such as par, txt, etc.

    Reference: `https://github.com/radis/radis/issues/121`
    """

    sf = SpectrumFactory(wavenum_min=2000, wavenum_max=3000, pressure=1)

    file_dir = getTestFile("cdsd_hitemp_09_fragment.txt")
    test_file = file_dir[:-8] + "*"
    sf.load_databank(path=test_file, format="cdsd-hitemp", parfuncfmt="hapi")

    try:
        sf.load_databank(path=test_file, format="cdsd-hitemp", parfuncfmt="hapi")
    except UnicodeDecodeError as err:
        raise UnicodeDecodeError(
            "Couldn't load database the 2nd time. This may be due to cache files trying to be read as normal files"
        ) from err


@pytest.mark.fast
def test_ignore_irrelevant_files(*args, **kwargs):
    """
    Implemented in https://github.com/radis/radis/pull/185

    - Test that calculating with cache files outside the spectral range they
    are properly ignored
    """

    from radis.misc.warning import EmptyDatabaseError, IrrelevantFileWarning

    # Regenerate .h5 cache file
    sf = SpectrumFactory(wavenum_min=2280, wavenum_max=2290)
    test_file = getTestFile("cdsd_hitemp_09_fragment.txt")
    sf.load_databank(
        path=test_file, format="cdsd-hitemp", parfuncfmt="hapi", db_use_cached="regen"
    )
    assert exists(test_file.replace(".txt", ".h5"))
    # Also note that there was no EmptyDatabaseError : file was properly loaded!

    # Load same .h5 cache file in another spectral range
    # ... Expect an error in use_cached = 'force' (file properly detected as
    # irrelevant)
    sf2 = SpectrumFactory(wavenum_min=100000, wavenum_max=100002)
    with pytest.raises(IrrelevantFileWarning):
        sf2.load_databank(
            path=test_file,
            format="cdsd-hitemp",
            parfuncfmt="hapi",
            db_use_cached="force",
        )

    # Again without 'force'
    # ... Expect no IrrelevantFile error, however range should be empty.
    with pytest.raises(EmptyDatabaseError):
        sf3 = SpectrumFactory(wavenum_min=100000, wavenum_max=100002)
        sf3.load_databank(path=test_file, format="cdsd-hitemp", parfuncfmt="hapi")


@pytest.mark.fast
def test_custom_abundance(verbose=True, plot=False, *args, **kwargs):
    """Test calculatinos with custom abundances

    set the terrestrial abundance to False to
    force computation of linestrength from Einstein coefficients instead
    of rescaling parameters

    :py:meth:`~radis.lbl.base.DatabankLoader.get_abundance`,
    :py:meth:`~radis.lbl.base.DatabankLoader.set_abundance`
    """

    sf = SpectrumFactory(
        2284.2,
        2284.6,
        wstep=0.001,  # cm-1
        pressure=20 * 1e-3,  # bar
        cutoff=0,
        path_length=0.1,
        mole_fraction=400e-6,
        molecule="CO2",
        isotope="1,2",
        medium="vacuum",
        truncation=5,
        verbose=0,
    )
    sf.warnings["MissingSelfBroadeningWarning"] = "ignore"
    sf.load_databank("HITEMP-CO2-TEST")

    #%% Compute a spectrum ; set the terrestrial abundance to False to
    # force computation of linestrength from Einstein coefficients instead
    # of rescaling parameters
    s = sf.eq_spectrum(2000)
    sf.molparam.terrestrial_abundances = False
    s2 = sf.eq_spectrum(2000)
    assert (
        s != s2
    )  # compare spectra; we should see small differences due to different methods
    assert s.compare_with(s2, "abscoeff", rtol=0.5e-2, plot=False)
    # assert np.allclose(2 * s2.get("abscoeff")[1], s.get("abscoeff")[1])

    if plot:
        from radis import plot_diff

        plot_diff(s, s2)

    #%% Now restart by changing the abundance itself ; as a user would do

    # ... check abundance get/set functions work well
    abundance12_0 = sf.get_abundance("CO2", [1, 2])

    if verbose:
        print("Abundance: CO2[1]", sf.get_abundance("CO2", 1))
        print("Abundance: CO2[1,2]", sf.get_abundance("CO2", [1, 2]))

    sf.set_abundance("CO2", [1, 2], abundance12_0 / 3)
    assert all(sf.get_abundance("CO2", [1, 2]) == abundance12_0 / 3)
    sf.set_abundance("CO2", 1, abundance12_0[0] / 3)
    sf.set_abundance("CO2", 2, abundance12_0[1] / 3)
    assert all(sf.get_abundance("CO2", [1, 2]) == abundance12_0 / 3)

    # Below we compute a spectrum with the abundance divided by 3; and then
    # check that the resulting absorption coefficient is 3x smaller

    s3 = sf.eq_spectrum(2000)
    assert s.compare_with((3 * s3.take("abscoeff")), "abscoeff", rtol=0.5e-2, plot=True)


def _run_testcases(verbose=True, plot=False):

    test_retrieve_from_database(plot=plot, verbose=verbose)
    test_ignore_cached_files()
    test_ignore_irrelevant_files(verbose=verbose)
    test_custom_abundance()


if __name__ == "__main__":
    _run_testcases()
